@@ -1,7 +1,7 @@
 {-|
   Copyright   :  (C) 2015-2016, University of Twente,
                      2017-2018, Google Inc.,
-                     2021-2023, QBayLogic B.V.,
+                     2021-2024, QBayLogic B.V.,
                      2022     , Google Inc.
   License     :  BSD2 (see the file LICENSE)
   Maintainer  :  QBayLogic B.V. <devops@qbaylogic.com>
@@ -31,6 +31,8 @@ import           Data.HashMap.Lazy                    (HashMap)
 import qualified Data.HashMap.Lazy                    as HashMap
 import           Data.HashSet                         (HashSet)
 import qualified Data.HashSet                         as HashSet
+import qualified Data.Map                             as Map
+import           Data.Map                             (Map)
 import           Data.List                            (nub, nubBy)
 import           Data.List.Extra                      ((<:>), zipEqual)
 import           Data.Maybe                           (catMaybes,fromMaybe,mapMaybe)
@@ -40,6 +42,7 @@ import qualified Data.Text.Lazy                       as Text
 import qualified Data.Text                            as TextS
 import           Data.Text.Prettyprint.Doc.Extra
 import qualified Data.Text.Prettyprint.Doc.Extra      as PP
+import           Data.Tuple                           (swap)
 import qualified System.FilePath
 
 import           Clash.Annotations.Primitive          (HDL (..))
@@ -59,13 +62,13 @@ import           Clash.Driver.Types                   (ClashOpts(..))
 import           Clash.Explicit.BlockRam.Internal     (unpackNats)
 import           Clash.Netlist.BlackBox.Types         (HdlSyn (..))
 import           Clash.Netlist.BlackBox.Util
-  (extractLiterals, renderBlackBox, renderFilePath)
+  (extractLiterals, renderBlackBox, addFilePath)
 import qualified Clash.Netlist.Id                     as Id
 import           Clash.Netlist.Types                  hiding (intWidth, usages, _usages)
 import           Clash.Netlist.Util
 import           Clash.Signal.Internal                (ActiveEdge (..))
 import           Clash.Util
-  (SrcSpan, noSrcSpan, curLoc, makeCached, indexNote)
+  (SrcSpan, clogBase, noSrcSpan, curLoc, makeCached, indexNote)
 import           Clash.Util.Graph                     (reverseTopSort)
 
 -- | State for the 'Clash.Backend.SystemVerilog.SystemVerilogM' monad:
@@ -82,8 +85,12 @@ data SystemVerilogState =
     , _includes  :: [(String,Doc)]
     , _imports   :: [Text.Text]
     , _libraries :: [Text.Text]
-    , _dataFiles      :: [(String,FilePath)]
-    -- ^ Files to be copied: (filename, old path)
+    , _dataFiles :: Map FilePath String
+    -- ^ Files to be copied: (path to file, file name). The file name stored in
+    -- the map might differ from the file name in the path, to prevent name
+    -- collisions.
+    , _includeDirs :: [FilePath]
+    -- ^ Import directories passed in with @-i@
     , _memoryDataFiles:: [(String,String)]
     -- ^ Files to be stored: (filename, contents). These files are generated
     -- during the execution of 'genNetlist'.
@@ -119,7 +126,8 @@ instance Backend SystemVerilogState where
     , _includes=[]
     , _imports=[]
     , _libraries=[]
-    , _dataFiles=[]
+    , _dataFiles=mempty
+    , _includeDirs = opt_importPaths opts
     , _memoryDataFiles=[]
     , _tyPkgCtx=False
     , _intWidth=opt_intWidth opts
@@ -188,10 +196,12 @@ instance Backend SystemVerilogState where
   addImports inps = imports %= (inps ++)
   addAndSetData f = do
     fs <- use dataFiles
-    let (fs',f') = renderFilePath fs f
+    iDirs <- use includeDirs
+    let (fs',f') = addFilePath iDirs fs f
     dataFiles .= fs'
     return f'
-  getDataFiles = use dataFiles
+  getDataFiles :: State SystemVerilogState [(String, FilePath)]
+  getDataFiles = map swap . Map.assocs <$> use dataFiles
   addMemoryDataFile f = memoryDataFiles %= (f:)
   getMemoryDataFiles = use memoryDataFiles
   ifThenElseExpr _ = True
@@ -1228,7 +1238,9 @@ expr_ _ (BlackBoxE pNm _ _ _ _ bbCtx _)
 expr_ _ (BlackBoxE pNm _ _ _ _ bbCtx _)
   | pNm == "Clash.Sized.Internal.Index.fromInteger#"
   , [Literal _ (NumLit n), Literal _ i] <- extractLiterals bbCtx
-  = exprLitSV (Just (Index (fromInteger n),fromInteger n)) i
+  , Just k <- clogBase 2 n
+  , let k' = max 1 k
+  = exprLitSV (Just (Index (fromInteger n),k')) i
 
 expr_ b (BlackBoxE _ libs imps inc bs bbCtx b') =
   parenIf (b || b') (Ap (renderBlackBox libs imps inc bs bbCtx <*> pure 0))
