@@ -1,7 +1,7 @@
 {-|
   Copyright   :  (C) 2015-2016, University of Twente,
                      2017-2018, Google Inc.,
-                     2021-2023, QBayLogic B.V.
+                     2021-2024, QBayLogic B.V.
                      2022     , Google Inc.
   License     :  BSD2 (see the file LICENSE)
   Maintainer  :  QBayLogic B.V. <devops@qbaylogic.com>
@@ -45,16 +45,22 @@ import           Data.HashMap.Strict                  (HashMap)
 import qualified Data.HashMap.Strict                  as HashMap
 import           Data.HashSet                         (HashSet)
 import qualified Data.HashSet                         as HashSet
+import qualified Data.Map                             as Map
+import           Data.Map                             (Map)
 import           Data.Maybe                           (catMaybes, fromMaybe)
 import           Data.Monoid                          (Ap(Ap))
 import           Data.Monoid.Extra                    ()
 import           Data.List
-  (mapAccumL, mapAccumR, nubBy, foldl')
+  (mapAccumL, mapAccumR, nubBy)
+#if !MIN_VERSION_base(4,20,0)
+import           Data.List                            (foldl')
+#endif
 import           Data.List.Extra                      ((<:>))
 import           Data.Text.Lazy                       (pack)
 import qualified Data.Text.Lazy                       as Text
 import qualified Data.Text                            as TextS
 import           Data.Text.Prettyprint.Doc.Extra
+import           Data.Tuple                           (swap)
 import qualified System.FilePath
 import           GHC.Stack                            (HasCallStack)
 
@@ -73,13 +79,13 @@ import           Clash.Driver.Types                   (ClashOpts(..))
 import           Clash.Explicit.BlockRam.Internal     (unpackNats)
 import           Clash.Netlist.BlackBox.Types         (HdlSyn)
 import           Clash.Netlist.BlackBox.Util
-  (extractLiterals, renderBlackBox, renderFilePath)
+  (extractLiterals, renderBlackBox, addFilePath)
 import qualified Clash.Netlist.Id                     as Id
 import           Clash.Netlist.Types as N             hiding (intWidth, usages, _usages)
 import           Clash.Netlist.Util
 import           Clash.Signal.Internal                (ActiveEdge (..))
 import           Clash.Util
-  (SrcSpan, noSrcSpan, curLoc, indexNote, makeCached)
+  (SrcSpan, clogBase, noSrcSpan, curLoc, indexNote, makeCached)
 
 -- | State for the 'Clash.Backend.Verilog.VerilogM' monad:
 data VerilogState =
@@ -91,8 +97,12 @@ data VerilogState =
     , _includes  :: [(String,Doc)]
     , _imports   :: HashSet Text.Text
     , _libraries :: HashSet Text.Text
-    , _dataFiles      :: [(String,FilePath)]
-    -- ^ Files to be copied: (filename, old path)
+    , _dataFiles :: Map FilePath String
+    -- ^ Files to be copied: (path to file, file name). The file name stored in
+    -- the map might differ from the file name in the path, to prevent name
+    -- collisions.
+    , _includeDirs :: [FilePath]
+    -- ^ Import directories passed in with @-i@
     , _memoryDataFiles:: [(String,String)]
     -- ^ Files to be stored: (filename, contents). These files are generated
     -- during the execution of 'genNetlist'.
@@ -123,7 +133,8 @@ instance Backend VerilogState where
     , _includes=[]
     , _imports=HashSet.empty
     , _libraries=HashSet.empty
-    , _dataFiles=[]
+    , _dataFiles=mempty
+    , _includeDirs = opt_importPaths opts
     , _memoryDataFiles=[]
     , _customConstrs=HashMap.empty
     , _intWidth=opt_intWidth opts
@@ -188,10 +199,11 @@ instance Backend VerilogState where
   addImports inps = imports %= (\s -> foldl' (flip HashSet.insert) s inps)
   addAndSetData f = do
     fs <- use dataFiles
-    let (fs',f') = renderFilePath fs f
+    iDirs <- use includeDirs
+    let (fs',f') = addFilePath iDirs fs f
     dataFiles .= fs'
     return f'
-  getDataFiles = use dataFiles
+  getDataFiles = map swap . Map.assocs <$> use dataFiles
   addMemoryDataFile f = memoryDataFiles %= (f:)
   getMemoryDataFiles = use memoryDataFiles
   ifThenElseExpr _ = True
@@ -1148,7 +1160,9 @@ expr_ _ (BlackBoxE pNm _ _ _ _ bbCtx _)
 expr_ _ (BlackBoxE pNm _ _ _ _ bbCtx _)
   | pNm == "Clash.Sized.Internal.Index.fromInteger#"
   , [Literal _ (NumLit n), Literal _ i] <- extractLiterals bbCtx
-  = exprLit undefValue (Just (Index (fromInteger n),fromInteger n)) i
+  , Just k <- clogBase 2 n
+  , let k' = max 1 k
+  = exprLitV (Just (Index (fromInteger n),k')) i
 
 expr_ b (BlackBoxE _ libs imps inc bs bbCtx b') = do
   parenIf (b || b') (Ap (renderBlackBox libs imps inc bs bbCtx <*> pure 0))
@@ -1213,20 +1227,27 @@ rtreeChain _                               = Nothing
 exprLitV :: Maybe (HWType,Size) -> Literal -> VerilogM Doc
 exprLitV = exprLit undefValue
 
-exprLit :: Lens' s (Maybe (Maybe Int)) -> Maybe (HWType,Size) -> Literal -> Ap (State s) Doc
+exprLit :: Backend s => Lens' s (Maybe (Maybe Int)) -> Maybe (HWType,Size) -> Literal -> Ap (State s) Doc
 exprLit _ Nothing (NumLit i) = integer i
 
-exprLit k (Just (hty,sz)) (NumLit i) = case hty of
+exprLit k (Just (hty,sz)) (NumLit i0) = case hty of
   Unsigned _
    | i < 0     -> string "-" <> int sz <> string "'d" <> integer (abs i)
    | otherwise -> int sz <> string "'d" <> integer i
-  Index _ -> int (typeSize hty) <> string "'d" <> integer i
+  Index n
+   | 0 <= i0 && i0 < n -> int (typeSize hty) <> string "'d" <> integer i0
+   | otherwise         -> hdlTypeErrValue hty
   Signed _
    | i < 0     -> string "-" <> int sz <> string "'sd" <> integer (abs i)
    | otherwise -> int sz <> string "'sd" <> integer i
   _ -> int sz <> string "'b" <> blit
   where
     blit = bits k (toBits sz i)
+    i = case hty of
+             Signed _ -> let mask = 2^(sz-1) in case divMod i0 mask of
+                (s,i'') | even s    -> i''
+                        | otherwise -> i'' - mask
+             _ -> i0 `mod` 2^sz
 exprLit k (Just (_,sz)) (BitVecLit m i) = int sz <> string "'b" <> bvlit
   where
     bvlit = bits k (toBits' sz m i)
