@@ -2,7 +2,7 @@
   Copyright   :  (C) 2012-2016, University of Twente,
                      2016-2017, Myrtle Software Ltd,
                      2017     , QBayLogic, Google Inc.
-                     2020-2023, QBayLogic,
+                     2020-2024, QBayLogic,
                      2022     , Google Inc.
 
   License     :  BSD2 (see the file LICENSE)
@@ -12,23 +12,23 @@
 -}
 
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveAnyClass #-}
 {-# LANGUAGE NamedFieldPuns #-}
 {-# LANGUAGE NondecreasingIndentation #-}
-{-# LANGUAGE QuasiQuotes #-}
-{-# LANGUAGE TemplateHaskell #-}
 {-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE QuasiQuotes #-}
 {-# LANGUAGE RecordWildCards #-}
+{-# LANGUAGE TemplateHaskell #-}
 
 module Clash.Driver where
 
 import           Control.Concurrent               (MVar, modifyMVar, modifyMVar_, newMVar, withMVar)
 import           Control.Concurrent.Async         (mapConcurrently_)
-import qualified Control.Concurrent.Supply        as Supply
 import           Control.DeepSeq
-import           Control.Exception                (throw)
+import           Control.Exception                (throw, Exception)
 import qualified Control.Monad                    as Monad
-import           Control.Monad                    (unless, foldM, forM, filterM)
-import           Control.Monad.Catch              (MonadMask)
+import           Control.Monad                    (unless, foldM, forM)
+import           Control.Monad.Catch              (MonadMask, MonadThrow (throwM))
 import           Control.Monad.Extra              (whenM, ifM, unlessM)
 import           Control.Monad.IO.Class           (MonadIO)
 import           Control.Monad.State              (evalState, get)
@@ -49,6 +49,8 @@ import qualified Data.HashSet                     as HashSet
 import           Data.Proxy                       (Proxy(..))
 import           Data.List                        (intercalate)
 import qualified Data.List                        as List
+import           Data.List.NonEmpty               (NonEmpty((:|)))
+import qualified Data.List.NonEmpty               as NonEmpty
 import           Data.Maybe                       (fromMaybe, maybeToList, mapMaybe)
 import qualified Data.Map.Ordered                 as OMap
 import           Data.Map.Ordered.Extra           ()
@@ -85,12 +87,10 @@ import           Text.Trifecta.Result
 
 #if MIN_VERSION_ghc(9,0,0)
 import           GHC.Builtin.Names                 (eqTyConKey, ipClassKey)
-import           GHC.Types.Unique                  (getKey)
 
 import           GHC.Types.SrcLoc                  (SrcSpan)
 #else
 import           PrelNames               (eqTyConKey, ipClassKey)
-import           Unique                  (getKey)
 
 import           SrcLoc                           (SrcSpan)
 #endif
@@ -144,12 +144,13 @@ import qualified Clash.Primitives.Verification    as P
 import qualified Clash.Primitives.Xilinx.ClockGen as P
 import           Clash.Primitives.Types
 import           Clash.Signal.Internal
-import           Clash.Unique                     (Unique, getUnique)
+import           Clash.Unique                     (Unique, getUnique, fromGhcUnique)
 import           Clash.Util
   (ClashException(..), reportTimeDiff,
    wantedLanguageExtensions, unwantedLanguageExtensions, curLoc)
 import           Clash.Util.Graph                 (reverseTopSort)
 import qualified Clash.Util.Interpolate           as I
+import qualified Clash.Util.Supply                as Supply
 
 -- | Worker function of 'splitTopEntityT'
 splitTopAnn
@@ -216,8 +217,8 @@ splitTopAnn tcm sp typ@(tyView -> FunTy {}) t@Synthesize{t_inputs} =
   --   * HasCallStack
   shouldNotHavePortName :: Type -> Bool
   shouldNotHavePortName (tyView -> TyConApp (nameUniq -> tcUniq) tcArgs)
-    | tcUniq == getKey eqTyConKey = True
-    | tcUniq == getKey ipClassKey
+    | tcUniq == fromGhcUnique eqTyConKey = True
+    | tcUniq == fromGhcUnique ipClassKey
     , [LitTy (SymTy "callStack"), _] <- tcArgs = True
   shouldNotHavePortName _ = False
 
@@ -475,7 +476,7 @@ generateHDL env design hdlState typeTrans peEval eval mainTopEntity startTime =
       -- TODO: Data files should go into their own directory
       -- FIXME: Files can silently overwrite each other
       hdlDocDigests <- mapM (writeHDL hdlDir) hdlDocs
-      dataFilesDigests <- copyDataFiles (opt_importPaths opts) hdlDir dfiles
+      dataFilesDigests <- copyDataFiles hdlDir dfiles
       memoryFilesDigests <- writeMemoryDataFiles hdlDir mfiles
 
       let
@@ -539,7 +540,7 @@ loadImportAndInterpret
   -- ^ Function name
   -> String
   -- ^ Type name ('BlackBoxFunction' or 'TemplateFunction')
-  -> m (Either Hint.InterpreterError a)
+  -> m (Either (NonEmpty Hint.InterpreterError) a)
 loadImportAndInterpret iPaths0 interpreterArgs topDir qualMod funcName typ = do
   Hint.liftIO $ Monad.when debugIsOn $
     putStr "Hint: Interpreting " >> putStrLn (qualMod ++ "." ++ funcName)
@@ -552,10 +553,10 @@ loadImportAndInterpret iPaths0 interpreterArgs topDir qualMod funcName typ = do
     Hint.unsafeInterpret funcName typ
 
   case bbfE of
-    Left _ -> do
+    Left globalException -> do
       -- Try to interpret module as a local module, not yet present in the
       -- global package database(s).
-      Hint.unsafeRunInterpreterWithArgsLibdir interpreterArgs topDir $ do
+      localRes <- Hint.unsafeRunInterpreterWithArgsLibdir interpreterArgs topDir $ do
         Hint.reset
         iPaths1 <- (iPaths0++) <$> Hint.get Hint.searchPath
         Hint.set [ Hint.searchPath Hint.:= iPaths1
@@ -563,8 +564,13 @@ loadImportAndInterpret iPaths0 interpreterArgs topDir qualMod funcName typ = do
         Hint.loadModules [qualMod]
         Hint.setImports [ "Clash.Netlist.BlackBox.Types", "Clash.Netlist.Types", qualMod]
         Hint.unsafeInterpret funcName typ
-    Right _ -> do
-      return bbfE
+
+      case localRes of
+        Left localException -> pure (Left (globalException :| [localException]))
+        Right res -> pure (Right res)
+
+    Right res -> do
+      return (Right res)
  where
    langExts = map Hint.asExtension $
                 map show wantedLanguageExtensions ++
@@ -631,11 +637,7 @@ compilePrimitive idirs pkgDbs topDir (BlackBoxHaskell bbName wf usedArgs multiRe
         let interpreterArgs = concatMap (("-package-db":) . (:[])) pkgDbs
         -- Compile a blackbox template function or fetch it from an already compiled file.
         r <- go interpreterArgs source
-        processHintError
-          (show bbGenName)
-          bbName
-          id
-          r
+        processHintErrors (show bbGenName) bbName r
 
   pure (BlackBoxHaskell bbName wf usedArgs multiRes bbGenName (hash source, bbFunc))
  where
@@ -656,7 +658,7 @@ compilePrimitive idirs pkgDbs topDir (BlackBoxHaskell bbName wf usedArgs multiRe
     go
       :: [String]
       -> Maybe Text
-      -> IO (Either Hint.InterpreterError BlackBoxFunction)
+      -> IO (Either (NonEmpty Hint.InterpreterError) BlackBoxFunction)
     go args (Just source') = do
       -- Create a temporary directory with user module in it, add it to the
       -- list of import direcotries, and run as if it were a "normal" compiled
@@ -709,7 +711,8 @@ compilePrimitive idirs pkgDbs topDir
       Text.writeFile (modDir </> last modNames <.>  "hs") source
       loadImportAndInterpret (tmpDir':idirs) iArgs topDir qualMod funcName "TemplateFunction"
     let hsh = hash (qualMod, source)
-    processHintError (show bbGenName) pNm (BBFunction (Data.Text.unpack pNm) hsh) r
+    BBFunction (Data.Text.unpack pNm) hsh <$>
+      processHintErrors (show bbGenName) pNm  r
   parseBB ((THaskell,bbGenName),Nothing) = do
     let BlackBoxFunctionName modNames funcName = bbGenName
         qualMod = intercalate "." modNames
@@ -720,36 +723,42 @@ compilePrimitive idirs pkgDbs topDir
         Just f -> pure f
         Nothing -> do
           r <- loadImportAndInterpret idirs iArgs topDir qualMod funcName "TemplateFunction"
-          processHintError (show bbGenName) pNm id r
+          processHintErrors (show bbGenName) pNm r
     pure (BBFunction (Data.Text.unpack pNm) hsh tf)
 
 compilePrimitive _ _ _ (Primitive pNm wf typ) =
   return (Primitive pNm wf typ)
 {-# SCC compilePrimitive #-}
 
-processHintError
-  :: Monad m
-  => String
-  -> Data.Text.Text
-  -> (t -> r)
-  -> Either Hint.InterpreterError t
-  -> m r
-processHintError fun bb go r = case r of
-  Left (Hint.GhcException err) ->
-    error' "GHC Exception" err
-  Left (Hint.NotAllowed err) ->
-    error' "NotAllowed error" err
-  Left (Hint.UnknownError err) ->
-    error' "an unknown error" err
-  Left (Hint.WontCompile ghcErrs) ->
-    error' "compilation errors" (intercalate "\n\n" $ map Hint.errMsg ghcErrs)
-  Right f ->
-    return $ go f
+newtype HintError = HintError String deriving (Exception)
+
+instance Show HintError where
+  showsPrec :: Int -> HintError -> ShowS
+  showsPrec _ (HintError e) = showString e
+
+processHintErrors ::
+  (MonadThrow m, Monad m) =>
+  -- | Function to interpret
+  String ->
+  -- | BlackBox function name
+  Data.Text.Text ->
+  -- | Hint result
+  Either (NonEmpty Hint.InterpreterError) t ->
+  m t
+processHintErrors fun bb r = case r of
+  Left es -> throwM $ HintError (formatExceptions (NonEmpty.toList es))
+  Right f -> pure f
  where
-  error' errType report =
-    error $ unwords [ "Encountered", errType, "while compiling blackbox template"
-                    , "function", show fun, "for function", show bb ++ "."
-                    , "Compilation reported: \n\n" ++ report ]
+  formatExceptions es = [I.i|
+    Encountered one or more exceptions when compiling blackbox template function
+    '#{fun}' for function '#{bb}'.
+  |] <> "\n\n" <> intercalate "\n\n" (map formatException es)
+
+  formatException e = [I.i|
+    Encountered:
+
+      #{e}
+  |]
 
 -- | Pretty print Components to HDL Documents
 createHDL
@@ -1011,17 +1020,16 @@ writeMemoryDataFiles dir files =
   forM files $ \(fname, content) ->
     writeAndHash (dir </> fname) (ByteStringLazyChar8.pack content)
 
--- | Copy data files added with ~FILE
+-- | Copy data files added with ~FILE, assumes the files are canonicalized already.
+-- Throws an error on relative paths as sanity check.
 copyDataFiles
-  :: [FilePath]
-  -- ^ Import directories passed in with @-i@
-  -> FilePath
+  :: FilePath
   -- ^ Directory to copy to
   -> [(FilePath,FilePath)]
   -- ^ [(name of newly made file in HDL output dir, file to copy)]
   -> IO [ByteString]
   -- ^ SHA256 hashes of written files
-copyDataFiles idirs targetDir = mapM copyDataFile
+copyDataFiles targetDir = mapM copyDataFile
  where
   copyDataFile :: (FilePath, FilePath) -> IO ByteString
   copyDataFile (newName, toCopy)
@@ -1030,28 +1038,7 @@ copyDataFiles idirs targetDir = mapM copyDataFile
         (doesFileExist toCopy)
         (copyAndHash toCopy (targetDir </> newName))
         (error [I.i|Could not find data file #{show toCopy}. Does it exist?|])
-    | otherwise = do
-      let candidates = map (</> toCopy) idirs
-      found <- filterM doesFileExist candidates
-      case found of
-        [] -> error [I.i|
-          Could not find data file #{show toCopy}. The following directories were
-          searched:
-
-            #{idirs}
-
-          You can add directories Clash will look in using `-i`.
-        |]
-        (_:_:_) -> error [I.i|
-          Multiple data files for #{show toCopy} found. The following candidates
-          were found:
-
-            #{found}
-
-          Please disambiguate data files.
-        |]
-        [c] ->
-          copyAndHash c (targetDir </> newName)
+    | otherwise = error [I.i|copyDataFiles received a relative path #{show toCopy}. This is a bug in Clash, please report it.|]
 
   copyAndHash src dst = do
     ifM
@@ -1120,8 +1107,8 @@ sortTop bindingsMap topEntities =
 
   mapFrom tops =
     let
-      topIndices = HashMap.fromList (zip (map topToUnique tops) [(0 :: Int)..])
+      topIndices = HashMap.fromList (zip (map topToUnique tops) [(0 :: Unique)..])
       nonOrdered = HashMap.fromListWith (<>) (map (second pure) edges)
-      orderFunc k = fromMaybe (-1) (HashMap.lookup k topIndices)
+      orderFunc k = HashMap.lookup k topIndices
     in
       HashMap.map (List.sortOn orderFunc) nonOrdered
