@@ -3,6 +3,7 @@
                     2017     , Myrtle Software Ltd,
                     2021-2022, QBayLogic B.V.
                     2022     , LUMI GUIDE FIETSDETECTIE B.V.
+                    2022     , Google Inc.
   License    :  BSD2 (see the file LICENSE)
   Maintainer :  QBayLogic B.V. <devops@qbaylogic.com>
 
@@ -18,6 +19,7 @@ module Clash.Netlist.BlackBox.Types
  , BlackBoxTemplate
  , TemplateKind (..)
  , Element(..)
+ , EscapedSymbol(..)
  , Decl(..)
  , HdlSyn(..)
  , RenderVoid(..)
@@ -34,7 +36,7 @@ import                GHC.Generics               (Generic)
 import                Clash.Core.Term            (Term)
 import                Clash.Core.Type            (Type)
 import {-# SOURCE #-} Clash.Netlist.Types
-  (BlackBox, NetlistMonad)
+  (BlackBox, NetlistMonad, Usage(Cont))
 
 import qualified      Clash.Signal.Internal      as Signal
 
@@ -52,10 +54,10 @@ data TemplateKind
   | TExpr
   deriving (Show, Eq, Generic, NFData, Binary, Hashable)
 
--- | See @Clash.Primitives.Types.BlackBox@ for documentation on this record's
+-- | See 'Clash.Primitives.Types.BlackBox' for documentation on this record's
 -- fields. (They are intentionally renamed to prevent name clashes.)
 data BlackBoxMeta =
-  BlackBoxMeta { bbOutputReg :: Bool
+  BlackBoxMeta { bbOutputUsage :: Usage
                , bbKind :: TemplateKind
                , bbLibrary :: [BlackBoxTemplate]
                , bbImports :: [BlackBoxTemplate]
@@ -67,9 +69,9 @@ data BlackBoxMeta =
                }
 
 -- | Use this value in your blackbox template function if you do want to
--- accept the defaults as documented in @Clash.Primitives.Types.BlackBox@.
+-- accept the defaults as documented in 'Clash.Primitives.Types.BlackBox'.
 emptyBlackBoxMeta :: BlackBoxMeta
-emptyBlackBoxMeta = BlackBoxMeta False TExpr [] [] [] [] NoRenderVoid [] []
+emptyBlackBoxMeta = BlackBoxMeta Cont TExpr [] [] [] [] NoRenderVoid [] []
 
 -- | A BlackBox function generates a blackbox template, given the inputs and
 -- result type of the function it should provide a blackbox for. This is useful
@@ -178,7 +180,8 @@ data Element
   -- ^ Period of a domain. Errors if not applied to a @KnownDomain@ or
   -- @KnownConfiguration@.
   | LongestPeriod
-  -- ^ Longest period of all known domains
+  -- ^ Longest period of all known domains. The minimum duration returned is
+  -- 100 ns, see https://github.com/clash-lang/clash-compiler/issues/2455.
   | ActiveEdge !Signal.ActiveEdge !Int
   -- ^ Test active edge of memory elements in a certain domain. Errors if not
   -- applied to a @KnownDomain@ or @KnownConfiguration@.
@@ -198,7 +201,7 @@ data Element
   -- always return 0 (False) if `-fclash-aggressive-x-optimization-blackboxes`
   -- is NOT set.
   | StrCmp [Element] !Int
-  | OutputWireReg !Int
+  | OutputUsage !Int
   | Vars !Int
   | GenSym [Element] !Int
   | Repeat [Element] [Element]
@@ -209,6 +212,12 @@ data Element
   | CtxName
   -- ^ The "context name", name set by `Clash.Magic.setName`, defaults to the
   -- name of the closest binder
+  | EscapedSymbol EscapedSymbol
+  -- ^ Used for "[\" and "\]", they'll be rendered as "[" and "]",
+  -- but pretty printed as "[\" and "\]".
+  deriving (Show, Generic, NFData, Binary, Eq, Hashable)
+
+data EscapedSymbol = SquareBracketOpen | SquareBracketClose
   deriving (Show, Generic, NFData, Binary, Eq, Hashable)
 
 -- | Component instantiation hole. First argument indicates which function argument
