@@ -1,8 +1,9 @@
 {-|
   Copyright  :  (C) 2012-2016, University of Twente,
                     2016-2017, Myrtle Software Ltd,
-                    2021-2022, QBayLogic B.V.
+                    2021-2023, QBayLogic B.V.
                     2022     , LUMI GUIDE FIETSDETECTIE B.V.
+                    2022     , Google Inc.
   License    :  BSD2 (see the file LICENSE)
   Maintainer :  QBayLogic B.V. <devops@qbaylogic.com>
 
@@ -13,17 +14,34 @@
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE LambdaCase #-}
+{-# LANGUAGE NumericUnderscores #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE QuasiQuotes #-}
 {-# LANGUAGE RankNTypes #-}
 {-# LANGUAGE TemplateHaskell #-}
 
-module Clash.Netlist.BlackBox.Util where
+module Clash.Netlist.BlackBox.Util
+    ( renderTemplate
+    , walkElement
+    , verifyBlackBoxContext
+    , onBlackBox
+    , setSym
+    , extractLiterals
+    , renderBlackBox
+    , getUsedArguments
+    , canonicalizeDataFilePath
+    , addFilePath
+    , exprToString
+    , renderElem
+    , getDomainConf
+    , prettyBlackBox
+    ) where
 
 import           Control.Exception               (throw)
 import           Control.Lens
   (use, (%=), _1, _2, element, (^?))
-import           Control.Monad                   (forM, (<=<))
+import           Control.Monad                   (forM, (<=<), filterM)
+import           Control.Monad.Extra             (ifM)
 import           Control.Monad.State             (State, StateT (..), lift, gets)
 import           Data.Bitraversable              (bitraverse)
 import           Data.Bool                       (bool)
@@ -34,11 +52,17 @@ import qualified Data.HashMap.Strict             as HashMap
 import qualified Data.IntMap                     as IntMap
 import           Data.List                       (nub)
 import           Data.List.Extra                 (indexMaybe)
+import qualified Data.Map                        as Map
+import           Data.Map                        (Map)
 import           Data.Maybe                      (mapMaybe, maybeToList, fromJust)
 import           Data.Monoid                     (Ap(getAp))
+import qualified Data.Set                        as Set
+import           Data.Set                        (Set)
 import qualified Data.Text
 import           Data.Text.Lazy                  (Text)
 import qualified Data.Text.Lazy                  as Text
+import           System.Directory                (doesFileExist, canonicalizePath)
+import           System.IO.Unsafe                (unsafePerformIO)
 
 #if MIN_VERSION_prettyprinter(1,7,0)
 import qualified Prettyprinter                   as PP
@@ -49,13 +73,13 @@ import qualified Data.Text.Prettyprint.Doc       as PP
 import           Data.Text.Prettyprint.Doc.Extra
 import           GHC.Stack                       (HasCallStack)
 import           System.FilePath                 (replaceBaseName, takeBaseName,
-                                                  takeFileName, (<.>))
+                                                  takeFileName, (<.>), isAbsolute, (</>))
 import           Text.Printf
 import           Text.Read                       (readEither)
 import           Text.Trifecta.Result            hiding (Err)
 
 import           Clash.Backend
-  (Backend (..), Usage (..), AggressiveXOptBB(..), RenderEnums(..))
+  (Backend (..), DomainMap, Usage (..), AggressiveXOptBB(..), RenderEnums(..))
 import           Clash.Netlist.BlackBox.Parser
 import           Clash.Netlist.BlackBox.Types
 import           Clash.Netlist.Types
@@ -63,7 +87,7 @@ import           Clash.Netlist.Types
    Declaration(BlackBoxD))
 import qualified Clash.Netlist.Id                as Id
 import qualified Clash.Netlist.Types             as N
-import           Clash.Netlist.Util              (typeSize, isVoid, stripVoid)
+import           Clash.Netlist.Util              (typeSize, isVoid, stripAttributes, stripVoid)
 import           Clash.Signal.Internal
   (ResetKind(..), ResetPolarity(..), InitBehavior(..), VDomainConfiguration (..))
 import           Clash.Util
@@ -117,15 +141,16 @@ inputHole = \case
   IsActiveEnable n -> pure n
   IsUndefined n    -> pure n
   StrCmp _ n       -> pure n
-  OutputWireReg n  -> pure n
+  OutputUsage n    -> pure n
   Vars n           -> pure n
   GenSym _ _       -> Nothing
   Repeat _ _       -> Nothing
   DevNull _        -> Nothing
   SigD _ nM        -> nM
   CtxName          -> Nothing
+  EscapedSymbol _  -> Nothing
 
--- | Determine if the number of normal/literal/function inputs of a blackbox
+-- | Determine if the number of normal\/literal\/function inputs of a blackbox
 -- context at least matches the number of argument that is expected by the
 -- template.
 verifyBlackBoxContext
@@ -183,10 +208,11 @@ verifyBlackBoxContext bbCtx (N.BBTemplate t) =
             Just n ->
               case indexMaybe (bbInputs bbCtx) n of
                 Just _ -> Nothing
-                Nothing ->
-                  Just ( "Blackbox required at least " ++ show (n+1)
-                      ++ " arguments, but only " ++ show (length (bbInputs bbCtx))
-                      ++ " were passed." )
+                Nothing -> do
+                  let str = fromJust (fmap Text.unpack (getAp $ prettyElem e))
+                  Just ( "Blackbox used \"" ++ str ++ "\""
+                      ++ ", but only " ++ show (length (bbInputs bbCtx))
+                      ++ " arguments were passed." )
 
 extractLiterals :: BlackBoxContext
                 -> [Expr]
@@ -227,7 +253,7 @@ setSym bbCtx l = do
               let decls = case typeSize hwTy of
                     0 -> []
                     _ -> [N.NetDecl Nothing nm' hwTy
-                         ,N.Assignment nm' e'
+                         ,N.Assignment nm' N.Cont e' -- TODO De-hardcode Cont
                          ]
               _2 %= (IntMap.insert i (Id.toText nm',decls))
               return (ToVar [Text (Id.toLazyText nm')] i)
@@ -289,22 +315,84 @@ setSym bbCtx l = do
                   ++ "blackbox for " ++ bbnm
         )
 
-selectNewName
-    :: Foldable t
-    => t String
-    -- ^ Set of existing names
-    -> FilePath
-    -- ^ Name for new file (
-    -> String
+type FileName = FilePath
+
+-- | Canonicalizes a datafile path, keeping in mind that the path might be
+-- relative to one of the import directories passed in with @-i@.
+canonicalizeDataFilePath ::
+  -- | Import directories passed in with @-i@
+  [FilePath] ->
+  -- | File to canonicalize
+  FilePath ->
+  FilePath
+canonicalizeDataFilePath idirs toCanonicalize
+  | isAbsolute toCanonicalize = unsafePerformIO $
+    ifM
+      (doesFileExist toCanonicalize)
+      (canonicalizePath toCanonicalize)
+      (error [I.i|Could not find data file #{show toCanonicalize}. Does it exist?|])
+  | otherwise = unsafePerformIO $ do
+      let candidates = map (</> toCanonicalize) idirs
+      found <- filterM doesFileExist candidates
+      case found of
+        [] -> error [I.i|
+          Could not find data file #{show toCanonicalize}. The following directories were
+          searched:
+            #{idirs}
+        |]
+        (_:_:_) -> error [I.i|
+          Multiple data files for #{show toCanonicalize} found. The following candidates
+          were found:
+            #{found}
+          Please disambiguate data files.
+        |]
+        [c] ->
+          canonicalizePath c
+{-# NOINLINE canonicalizeDataFilePath #-} -- To contain unsafePerformIO
+
+-- | Select a new file name that doesn't collide with existing names. Might return
+-- the same name as given if there are no collisions.
+selectNewName ::
+  -- | Set of existing names
+  Set FileName ->
+  -- | Name for file
+  FileName ->
+  -- | Updated file name
+  FileName
 selectNewName as a
-  | elem a as = selectNewName as (replaceBaseName a (takeBaseName a ++ "_"))
+  | a `Set.member` as = selectNewName as (replaceBaseName a (takeBaseName a ++ "_"))
   | otherwise = a
 
-renderFilePath :: [(String,FilePath)] -> String -> ([(String,FilePath)],String)
-renderFilePath fs f = ((f'',f):fs, f'')
-  where
-    f'  = takeFileName f
-    f'' = selectNewName (map fst fs) f'
+-- | Add a data file to a collection of existing files. Typically used in the HDL
+-- backend's 'addAndSetData'.
+addFilePath ::
+  -- | Import directories passed in with @-i@
+  [FilePath] ->
+  -- | Map of existing file paths to file names. The file names are guaranteed
+  -- to be unique, as long as this function is used to add new file paths.
+  Map FilePath FileName ->
+  -- | New file path to add. The canonicalized version of this file path is stored
+  -- in the map. See 'canonicalizeFilePath'.
+  FilePath ->
+  -- | Updated map and filename. Note that the returned filename might be different
+  -- from the filename in the given filepath in case of naming collisions.
+  (Map FilePath FileName, FileName)
+addFilePath iDirs existing path =
+  case Map.lookup canonicalizedPath existing of
+    Just existingName ->
+      -- This path was already added, return existing map and name
+      (existing, existingName)
+    Nothing ->
+      -- Select a new name that doesn't collide with existing names
+      ( Map.insert canonicalizedPath newFileName existing
+      , newFileName
+      )
+ where
+  canonicalizedPath = canonicalizeDataFilePath iDirs path
+
+  allNames = Set.fromList (Map.elems existing)
+  fileName = takeFileName canonicalizedPath
+  newFileName = selectNewName allNames fileName
 
 -- | Render a blackbox given a certain context. Returns a filled out template
 -- and a list of 'hidden' inputs that must be added to the encompassing component.
@@ -392,50 +480,56 @@ renderElem
 renderElem b (Component (Decl n subN (l:ls))) = do
   (o,oTy,_) <- idToExpr <$> bitraverse (lineToIdentifier b) (return . lineToType b) l
   is <- mapM (fmap idToExpr . bitraverse (lineToIdentifier b) (return . lineToType b)) ls
+  sp <- getSrcSpan
   let func0 = IntMap.lookup n (bbFunctions b)
       errr = concat [ "renderElem: not enough functions rendered? Needed "
                     , show (subN +1 ), " got only ", show (length (fromJust func0)) ]
-      func1 = indexNote' errr subN <$> func0
-      Just (templ0,_,libs,imps,inc,pCtx) = func1
-      b' = pCtx { bbResults = [(o,oTy)], bbInputs = bbInputs pCtx ++ is }
-      layoutOptions = LayoutOptions (AvailablePerLine 120 0.4)
-      render = N.BBTemplate . parseFail . renderLazy . layoutPretty layoutOptions
+  case indexNote' errr subN <$> func0 of
+    Just (templ0,_,libs,imps,inc,pCtx) -> do
+      let b' = pCtx { bbResults = [(o,oTy)], bbInputs = bbInputs pCtx ++ is }
+          layoutOptions = LayoutOptions (AvailablePerLine 120 0.4)
+          render = N.BBTemplate . parseFail b' . renderLazy . layoutPretty layoutOptions
 
-  templ1 <-
-    case templ0 of
-      Left t ->
-        return t
-      Right (nm0,ds) -> do
-        nm1 <- Id.next nm0
-        block <- getAp (blockDecl nm1 ds)
-        return (render block)
-
-  templ4 <-
-    case templ1 of
-      N.BBFunction {} ->
-        return templ1
-      N.BBTemplate templ2 -> do
-        (templ3, templDecls) <- setSym b' templ2
-        case templDecls of
-          [] ->
-            return (N.BBTemplate templ3)
-          _ -> do
-            nm1 <- Id.toText <$> Id.makeBasic "bb"
-            nm2 <- Id.makeBasic "bb"
-            let bbD = BlackBoxD nm1 libs imps inc (N.BBTemplate templ3) b'
-            block <- getAp (blockDecl nm2 (templDecls ++ [bbD]))
+      templ1 <-
+        case templ0 of
+          Left t ->
+            return t
+          Right (nm0,ds) -> do
+            nm1 <- Id.next nm0
+            block <- getAp (blockDecl nm1 ds)
             return (render block)
 
-  case verifyBlackBoxContext b' templ4 of
-    Nothing -> do
-      bb <- renderBlackBox libs imps inc templ4 b'
-      return (renderLazy . layoutPretty layoutOptions . bb)
-    Just err0 -> do
-      sp <- getSrcSpan
-      let err1 = concat [ "Couldn't instantiate blackbox for "
-                        , Data.Text.unpack (bbName b), ". Verification procedure "
-                        , "reported:\n\n" ++ err0 ]
-      throw (ClashException sp ($(curLoc) ++ err1) Nothing)
+      templ4 <-
+        case templ1 of
+          N.BBFunction {} ->
+            return templ1
+          N.BBTemplate templ2 -> do
+            (templ3, templDecls) <- setSym b' templ2
+            case templDecls of
+              [] ->
+                return (N.BBTemplate templ3)
+              _ -> do
+                nm1 <- Id.toText <$> Id.makeBasic "bb"
+                nm2 <- Id.makeBasic "bb"
+                let bbD = BlackBoxD nm1 libs imps inc (N.BBTemplate templ3) b'
+                block <- getAp (blockDecl nm2 (templDecls ++ [bbD]))
+                return (render block)
+
+      case verifyBlackBoxContext b' templ4 of
+        Nothing -> do
+          bb <- renderBlackBox libs imps inc templ4 b'
+          return (renderLazy . layoutPretty layoutOptions . bb)
+        Just err0 -> do
+          let err1 = concat [ "Couldn't instantiate blackbox for "
+                            , Data.Text.unpack (bbName b), ". Verification procedure "
+                            , "reported:\n\n" ++ err0 ]
+          throw (ClashException sp ($(curLoc) ++ err1) Nothing)
+    Nothing ->
+      let err1 = concat [show n
+                        , "'th argument isn't a function, only "
+                        , show (IntMap.keys (bbFunctions b))
+                        , "are."]
+       in throw (ClashException sp ($(curLoc) ++ err1) Nothing)
 
 renderElem b (SigD e m) = do
   e' <- Text.concat <$> mapM (fmap ($ 0) . renderElem b) e
@@ -456,7 +550,9 @@ renderElem b (Period n) = do
 
 renderElem _ LongestPeriod = do
   doms <- domainConfigurations
-  let longestPeriod = maximum [vPeriod v | v <- HashMap.elems doms]
+  -- Longest period with a minimum of 100 ns, see:
+  -- https://github.com/clash-lang/clash-compiler/issues/2455
+  let longestPeriod = maximum (100_000 : [vPeriod v | v <- HashMap.elems doms])
   return (const (Text.pack (show longestPeriod)))
 
 renderElem b (Tag n) = do
@@ -466,6 +562,8 @@ renderElem b (Tag n) = do
       return (const (Text.pack (Data.Text.unpack dom)))
     Clock dom ->
       return (const (Text.pack (Data.Text.unpack dom)))
+    ClockN dom ->
+      return (const (Text.pack (Data.Text.unpack dom)))
     Reset dom ->
       return (const (Text.pack (Data.Text.unpack dom)))
     Enable dom ->
@@ -480,45 +578,57 @@ renderElem b (IF c t f) = do
   syn <- hdlSyn
   enums <- renderEnums
   xOpt <- aggressiveXOptBB
-  let c' = check (coerce xOpt) iw hdl syn enums c
+  c' <- check (coerce xOpt) iw hdl syn enums c
   if c' > 0 then renderTemplate b t else renderTemplate b f
   where
-    check :: Bool -> Int -> HDL -> HdlSyn -> RenderEnums -> Element -> Int
+    check :: Backend backend => Bool -> Int -> HDL -> HdlSyn -> RenderEnums -> Element -> State backend Int
     check xOpt iw hdl syn enums c' = case c' of
-      (Size e)   -> typeSize (lineToType b [e])
-      (Length e) -> case lineToType b [e] of
+      (Size e)   -> pure $ typeSize (lineToType b [e])
+      (Length e) -> pure $ case lineToType b [e] of
                        (Vector n _)              -> n
                        Void (Just (Vector n _))  -> n
                        (MemBlob n _)             -> n
                        Void (Just (MemBlob n _)) -> n
                        _                         -> 0 -- HACK: So we can test in splitAt if one of the
                               -- vectors in the tuple had a zero length
-      (Lit n) -> case bbInputs b !! n of
+      (Lit n) -> pure $ case bbInputs b !! n of
         (l,_,_)
           | Literal _ l' <- l ->
             case l' of
+              -- Integer, Int#, KnownNat, Natural, Word#
               NumLit i -> fromInteger i
+              -- Bit
               BitLit bl -> case bl of
                 N.H -> 1
                 N.L -> 0
                 _   -> error $ $(curLoc) ++ "IF: LIT bit literal must be high or low"
+              -- Bool
               BoolLit bl -> bool 0 1 bl
               _ -> error $ $(curLoc) ++ "IF: LIT must be a numeric lit"
+          -- Int
           | DataCon (Signed _) _ [Literal _ (NumLit i)] <- l
             -> fromInteger i
+          -- Word, SNat
           | DataCon (Unsigned _) _ [Literal _ (NumLit i)] <- l
             -> fromInteger i
+          | BlackBoxE pNm _lib _use _incl _templ bbCtx _paren <- l
+          , pNm `elem` ["GHC.Int.I8#", "GHC.Int.I16#", "GHC.Int.I32#", "GHC.Int.I64#"
+                       ,"GHC.Word.W8#","GHC.Word.W16#","GHC.Word.W32#","GHC.Word.W64#"
+                       ,"GHC.Types.I#","GHC.Types.W#"
+                       ]
+          , [Literal _ (NumLit j)] <- extractLiterals bbCtx
+          -> fromInteger j
         k -> error $ $(curLoc) ++ ("IF: LIT must be a numeric lit:" ++ show k)
-      (Depth e)  -> case lineToType b [e] of
+      (Depth e)  -> pure $ case lineToType b [e] of
                       (RTree n _) -> n
                       _ -> error $ $(curLoc) ++ "IF: treedepth of non-tree type"
-      IW64       -> if iw == 64 then 1 else 0
-      (HdlSyn s) -> if s == syn then 1 else 0
-      (IsVar n)  -> let (e,_,_) = bbInputs b !! n
+      IW64       -> pure $ if iw == 64 then 1 else 0
+      (HdlSyn s) -> pure $ if s == syn then 1 else 0
+      (IsVar n)  -> pure $ let (e,_,_) = bbInputs b !! n
                     in case e of
                       Identifier _ Nothing -> 1
                       _                    -> 0
-      (IsLit n)  -> let (e,_,_) = bbInputs b !! n
+      (IsLit n)  -> pure $ let (e,_,_) = bbInputs b !! n
                     in case e of
                       DataCon {}   -> 1
                       Literal {}   -> 1
@@ -532,13 +642,13 @@ renderElem b (IF c t f) = do
                                                        RenderEnums True  -> 1
                                                        RenderEnums False -> 0
                           isScalar _ _            = 0
-                      in isScalar hdl ty
+                        in pure $ isScalar hdl ty
 
-      (IsUndefined n) ->
+      (IsUndefined n) -> pure $
         let (e, _, _) = bbInputs b !! n
         in if xOpt && checkUndefined e then 1 else 0
 
-      (IsActiveEnable n) ->
+      (IsActiveEnable n) -> pure $
         let (e, ty, _) = bbInputs b !! n in
         case ty of
           Enable _ ->
@@ -558,58 +668,92 @@ renderElem b (IF c t f) = do
               Literal Nothing (BoolLit False) -> 1
               _                               -> 1
           _ ->
-            error $ $(curLoc) ++ "IsActiveEnable: Expected Bool, not: " ++ show ty
+            error $ $(curLoc) ++ "IsActiveEnable: Expected Bool or Enable, not: " ++ show ty
 
-      (ActiveEdge edgeRequested n) ->
-        let (_, ty, _) = bbInputs b !! n in
-        case stripVoid ty of
-          KnownDomain _ _ edgeActual _ _ _ ->
+      (ActiveEdge edgeRequested n) -> do
+        let (_, ty, _) = bbInputs b !! n
+        domConf <- getDomainConf ty
+        case domConf of
+          VDomainConfiguration _ _ edgeActual _ _ _ -> pure $
             if edgeRequested == edgeActual then 1 else 0
-          _ ->
-            error $ $(curLoc) ++ "ActiveEdge: Expected `KnownDomain` or `KnownConfiguration`, not: " ++ show ty
 
-      (IsSync n) ->
-        let (_, ty, _) = bbInputs b !! n in
-        case stripVoid ty of
-          KnownDomain _ _ _ Synchronous _ _ -> 1
-          KnownDomain _ _ _ Asynchronous _ _ -> 0
-          _ -> error $ $(curLoc) ++ "IsSync: Expected `KnownDomain` or `KnownConfiguration`, not: " ++ show ty
+      (IsSync n) -> do
+        let (_, ty, _) = bbInputs b !! n
+        domConf <- getDomainConf ty
+        case domConf of
+          VDomainConfiguration _ _ _ Synchronous _ _ -> pure 1
+          VDomainConfiguration _ _ _ Asynchronous _ _ -> pure 0
 
-      (IsInitDefined n) ->
-        let (_, ty, _) = bbInputs b !! n in
-        case stripVoid ty of
-          KnownDomain _ _ _ _ Defined _ -> 1
-          KnownDomain _ _ _ _ Unknown _ -> 0
-          _ -> error $ $(curLoc) ++ "IsInitDefined: Expected `KnownDomain` or `KnownConfiguration`, not: " ++ show ty
+      (IsInitDefined n) -> do
+        let (_, ty, _) = bbInputs b !! n
+        domConf <- getDomainConf ty
+        case domConf of
+          VDomainConfiguration _ _ _ _ Defined _ -> pure 1
+          VDomainConfiguration _ _ _ _ Unknown _ -> pure 0
 
-      (IsActiveHigh n) ->
-        let (_, ty, _) = bbInputs b !! n in
-        case stripVoid ty of
-          KnownDomain _ _ _ _ _ ActiveHigh -> 1
-          KnownDomain _ _ _ _ _ ActiveLow -> 0
-          _ -> error $ $(curLoc) ++ "IsActiveHigh: Expected `KnownDomain` or `KnownConfiguration`, not: " ++ show ty
+      (IsActiveHigh n) -> do
+        let (_, ty, _) = bbInputs b !! n
+        domConf <- getDomainConf ty
+        case domConf of
+          VDomainConfiguration _ _ _ _ _ ActiveHigh -> pure 1
+          VDomainConfiguration _ _ _ _ _ ActiveLow  -> pure 0
 
-      (StrCmp [Text t1] n) ->
+      (StrCmp [Text t1] n) -> pure $
         let (e,_,_) = bbInputs b !! n
         in  case exprToString e of
               Just t2
                 | t1 == Text.pack t2 -> 1
                 | otherwise -> 0
               Nothing -> error $ $(curLoc) ++ "Expected a string literal: " ++ show e
-      (And es)   -> if all (/=0) (map (check xOpt iw hdl syn enums) es)
+      (And es)   -> do
+        es' <- mapM (check xOpt iw hdl syn enums) es
+        pure $ if all (/=0) es'
                        then 1
                        else 0
-      CmpLE e1 e2 -> if check xOpt iw hdl syn enums e1 <= check xOpt iw hdl syn enums e2
-                        then 1
-                        else 0
+      CmpLE e1 e2 -> do
+        v1 <- check xOpt iw hdl syn enums e1
+        v2 <- check xOpt iw hdl syn enums e2
+        if v1 <= v2
+          then pure 1
+          else pure 0
       _ -> error $ $(curLoc) ++ "IF: condition must be: SIZE, LENGTH, LIT, DEPTH, IW64, VIVADO, OTHERSYN, ISVAR, ISLIT, ISUNDEFINED, ISACTIVEENABLE, ACTIVEEDGE, ISSYNC, ISINITDEFINED, ISACTIVEHIGH, STRCMP, AND, ISSCALAR or CMPLE."
                              ++ "\nGot: " ++ show c'
 renderElem b e = fmap const (renderTag b e)
 
-parseFail :: Text -> BlackBoxTemplate
-parseFail t = case runParse t of
+getDomainConf :: (Backend backend, HasCallStack) => HWType -> State backend VDomainConfiguration
+getDomainConf = generalGetDomainConf domainConfigurations
+
+generalGetDomainConf
+  :: forall m. (Monad m, HasCallStack)
+  => (m DomainMap) -- ^ a way to get the `DomainMap`
+  -> HWType -> m VDomainConfiguration
+generalGetDomainConf getDomainMap ty = case (snd . stripAttributes . stripVoid) ty of
+  KnownDomain dom period activeEdge resetKind initBehavior resetPolarity ->
+    pure $ VDomainConfiguration (Data.Text.unpack dom) (fromIntegral period) activeEdge resetKind initBehavior resetPolarity
+
+  Clock dom -> go dom
+  ClockN dom -> go dom
+  Reset dom  -> go dom
+  Enable dom -> go dom
+  Product _DiffClock _ [Clock dom,_clkN] -> go dom
+  t -> error $ "Don't know how to get a Domain out of HWType: " <> show t
+ where
+  go :: HasCallStack => N.DomainName -> m VDomainConfiguration
+  go dom = do
+    doms <- getDomainMap
+    case HashMap.lookup dom doms of
+      Nothing -> error $ "Can't find domain " <> show dom <> ". Please report an issue at https://github.com/clash-lang/clash-compiler/issues."
+      Just conf -> pure conf
+
+parseFail :: BlackBoxContext -> Text -> BlackBoxTemplate
+parseFail b t = case runParse t of
   Failure errInfo ->
-    error (show (_errDoc errInfo))
+    error $ unlines
+        [ "error while parsing blackbox: " <> Data.Text.unpack (bbName b)
+        , "in component " <> Data.Text.unpack (Id.toText $ bbCompName b)
+        , "error:"
+        , show (_errDoc errInfo)
+        ]
   Success templ -> templ
 
 idToExpr
@@ -690,6 +834,7 @@ renderTag b (Lit n) =
 
   mkLit (BlackBoxE pNm _ _ _ _ bbCtx _) | pNm `elem` ["GHC.Int.I8#", "GHC.Int.I16#", "GHC.Int.I32#", "GHC.Int.I64#"
                                                      ,"GHC.Word.W8#","GHC.Word.W16#","GHC.Word.W32#","GHC.Word.W64#"
+                                                     ,"GHC.Types.I#","GHC.Types.W#"
                                                      ]
                                         , [Literal _ i] <- extractLiterals bbCtx
                                         = Literal Nothing i
@@ -789,9 +934,19 @@ renderTag b (FilePath e)    = case e of
 renderTag b (IncludeName n) = case indexMaybe (bbQsysIncName b) n of
   Just nm -> return (Text.fromStrict nm)
   _ -> error $ $(curLoc) ++ "~INCLUDENAME[" ++ show n ++ "] does not correspond to any index of the 'includes' field that is specified in the primitive definition"
-renderTag b (OutputWireReg n) = case IntMap.lookup n (bbFunctions b) of
-  Just ((_,rw,_,_,_,_):_) -> case rw of {N.Wire -> return "wire"; N.Reg -> return "reg"}
-  _ -> error $ $(curLoc) ++ "~OUTPUTWIREREG[" ++ show n ++ "] used where argument " ++ show n ++ " is not a function"
+renderTag b (OutputUsage n) = do
+  hdl <- gets hdlKind
+
+  let u = case IntMap.lookup n (bbFunctions b) of
+            Just ((_,u',_,_,_,_):_) -> u'
+            _ -> error $ $(curLoc) ++ "~OUTPUTUSAGE[" ++ show n ++ "] used where argument " ++ show n ++ " is not a function"
+
+  pure $ case (hdl, u) of
+    (VHDL, N.Proc N.Blocking) -> "variable"
+    (VHDL, _) -> "signal"
+
+    (_, N.Cont) -> "wire"
+    (_, _) -> "reg"
 renderTag b (Repeat [es] [i]) = do
   i'  <- Text.unpack <$> renderTag b i
   es' <- renderTag b es
@@ -839,6 +994,9 @@ renderTag b CtxName = case bbCtxName b of
     -> return (Id.toLazyText t)
   _ -> error "internal error"
 
+renderTag _ (EscapedSymbol sym) = case sym of
+  SquareBracketOpen  -> return "["
+  SquareBracketClose -> return "]"
 
 renderTag _ e = do e' <- getAp (prettyElem e)
                    error $ $(curLoc) ++ "Unable to evaluate: " ++ show e'
@@ -1033,7 +1191,7 @@ prettyElem (SigD es mI) = do
            (((string "~SIGD" <> brackets (string es')) <>) . int)
            mI)
 prettyElem (Vars i) = renderOneLine <$> (string "~VARS" <> brackets (int i))
-prettyElem (OutputWireReg i) = renderOneLine <$> (string "~RESULTWIREREG" <> brackets (int i))
+prettyElem (OutputUsage n) = renderOneLine <$> (string "~OUTPUTUSAGE" <> brackets (int n))
 prettyElem (ArgGen n x) =
   renderOneLine <$> (string "~ARGN" <> brackets (int n) <> brackets (int x))
 prettyElem (Template bbname source) = do
@@ -1043,6 +1201,9 @@ prettyElem (Template bbname source) = do
                                   <> brackets (string $ Text.concat bbname')
                                   <> brackets (string $ Text.concat source'))
 prettyElem CtxName = return "~CTXNAME"
+prettyElem (EscapedSymbol sym) = case sym of
+  SquareBracketOpen  -> return "[\\"
+  SquareBracketClose -> return "\\]"
 
 -- | Recursively walk @Element@, applying @f@ to each element in the tree.
 walkElement
@@ -1108,11 +1269,12 @@ walkElement f el = maybeToList (f el) ++ walked
         IsActiveEnable _ -> []
         IsUndefined _ -> []
         StrCmp es _ -> concatMap go es
-        OutputWireReg _ -> []
+        OutputUsage _ -> []
         Vars _ -> []
         Repeat es1 es2 ->
           concatMap go es1 ++ concatMap go es2
         CtxName -> []
+        EscapedSymbol _ -> []
 
 -- | Determine variables used in an expression. Used for VHDL sensitivity list.
 -- Also see: https://github.com/clash-lang/clash-compiler/issues/365
@@ -1186,7 +1348,7 @@ getUsedArguments (N.BBTemplate t) = nub (concatMap (walkElement matchArg) t)
         IW64 -> Nothing
         Length _ -> Nothing
         MaxIndex _ -> Nothing
-        OutputWireReg _ -> Nothing
+        OutputUsage _ -> Nothing
         Repeat _ _ -> Nothing
         Result -> Nothing
         Sel _ _ -> Nothing
@@ -1201,6 +1363,7 @@ getUsedArguments (N.BBTemplate t) = nub (concatMap (walkElement matchArg) t)
         TypM _ -> Nothing
         Vars _ -> Nothing
         CtxName -> Nothing
+        EscapedSymbol _ -> Nothing
 
 onBlackBox
   :: (BlackBoxTemplate -> r)
