@@ -2,7 +2,8 @@
   Copyright  :  (C) 2012-2016, University of Twente,
                     2017     , Myrtle Software Ltd,
                     2017-2018, Google Inc.
-                    2020-2022, QBayLogic B.V.
+                    2020-2023, QBayLogic B.V.
+                    2022-2023, Google Inc.
   License    :  BSD2 (see the file LICENSE)
   Maintainer :  QBayLogic B.V. <devops@qbaylogic.com>
 
@@ -38,6 +39,8 @@ import Control.Monad.Reader                 (ReaderT, MonadReader)
 import qualified Control.Monad.State        as Lazy (State)
 import qualified Control.Monad.State.Strict as Strict
   (State, MonadIO, MonadState, StateT)
+import Data.Aeson                           (FromJSON(..))
+import qualified Data.Aeson as Aeson
 import Data.Bits                            (testBit)
 import Data.Binary                          (Binary(..))
 import Data.Function                        (on)
@@ -48,6 +51,7 @@ import qualified Data.List                  as List
 import Data.IntMap                          (IntMap, empty)
 import Data.Map.Ordered                     (OMap)
 import Data.Map                             (Map)
+import qualified Data.Map as Map
 import Data.Maybe                           (mapMaybe)
 import Data.Monoid                          (Ap(..))
 import qualified Data.Set                   as Set
@@ -65,13 +69,14 @@ import GHC.Types.SrcLoc                     (SrcSpan)
 import SrcLoc                               (SrcSpan)
 #endif
 
+import Clash.Annotations.SynthesisAttributes(Attr)
 import Clash.Annotations.BitRepresentation  (FieldAnn)
 import Clash.Annotations.Primitive          (HDL(..))
 import Clash.Annotations.TopEntity          (TopEntity)
-import Clash.Backend                        (Backend)
+import Clash.Backend                        (Backend, HasUsageMap (..))
 import Clash.Core.HasType
 import Clash.Core.Type                      (Type)
-import Clash.Core.Var                       (Attr', Id)
+import Clash.Core.Var                       (Id)
 import Clash.Core.TyCon                     (TyConMap)
 import Clash.Core.VarEnv                    (VarEnv)
 import Clash.Driver.Types                   (BindingMap, ClashEnv(..), ClashOpts(..))
@@ -84,7 +89,7 @@ import Clash.Unique                         (Unique)
 import Clash.Annotations.BitRepresentation.Internal
   (CustomReprs, DataRepr', ConstrRepr')
 
-import {-# SOURCE #-} qualified Clash.Netlist.Id as Id
+import {-# SOURCE #-} qualified Clash.Netlist.Id as Id (toText)
 
 -- | Structure describing a top entity: it's id and its port annotations.
 data TopEntityT = TopEntityT
@@ -202,7 +207,7 @@ data IdentifierSet
 -- identifiers internally. Any Identifier should be trivially printable to any
 -- HDL.
 --
--- __NB__: use the functions in Clash.Netlist.Id. Don't use these constructors
+-- __NB__: use the functions in "Clash.Netlist.Id". Don't use these constructors
 -- directly.
 data Identifier
   -- | Unparsed identifier. Used for things such as port names, which should
@@ -236,7 +241,7 @@ data Identifier
     -- name of 'foo' and an extension of [6, 5] would render as 'foo_5_6'. Note
     -- that extensions are stored in reverse order for easier manipulation.
     , i_idType :: !IdentifierType
-    -- ^ See "IdentifierType".
+    -- ^ See 'IdentifierType'.
     , i_hdl :: !HDL
     -- ^ HDL this identifier is generated for.
     , i_provenance :: !CallStack
@@ -283,6 +288,7 @@ data ComponentMeta = ComponentMeta
   { cmWereVoids :: [Bool]
   , cmLoc :: SrcSpan
   , cmScope :: IdentifierSet
+  , cmUsage :: UsageMap
   } deriving (Generic, Show, NFData)
 
 type ComponentMap = OMap Unique (ComponentMeta, Component)
@@ -338,6 +344,10 @@ data NetlistState
   , _backend :: SomeBackend
   -- ^ The current HDL backend
   , _htyCache :: HWMap
+  , _usages :: UsageMap
+  -- ^ The current way signals are assigned in netlist. This is used to
+  -- determine how signals are rendered in HDL (i.e. wire/reg in Verilog, or
+  -- signal/variable in VHDL).
   }
 
 data ComponentPrefix
@@ -365,7 +375,7 @@ data Component
   = Component
   { componentName :: !Identifier -- ^ Name of the component
   , inputs        :: [(Identifier,HWType)] -- ^ Input ports
-  , outputs       :: [(WireOrReg,(Identifier,HWType),Maybe Expr)] -- ^ Output ports
+  , outputs       :: [(Usage,(Identifier,HWType),Maybe Expr)] -- ^ Output ports
   , declarations  :: [Declaration] -- ^ Internal declarations
   }
   deriving (Show, Generic, NFData)
@@ -373,11 +383,16 @@ data Component
 -- | Check if an input port is really an inout port.
 --
 isBiDirectional :: (Identifier, HWType) -> Bool
-isBiDirectional (_, BiDirectional _ _) = True
-isBiDirectional _ = False
+isBiDirectional = go . snd
+  where
+    go BiDirectional{} = True
+    go (Annotated _ hwty) = go hwty
+    go _ = False
 
 -- | Find the name and domain name of each clock argument of a component.
 --
+-- This will not consider @ClockN@ to be a clock argument, which means only the
+-- positive phase of a differential pair will be added to @sdcClock@.
 findClocks :: Component -> [(Text, Text)]
 findClocks (Component _ is _ _) =
   mapMaybe isClock is
@@ -437,6 +452,8 @@ data HWType
   -- ^ Sum-of-Product type: Name and Constructor names + field types
   | Clock !DomainName
   -- ^ Clock type corresponding to domain /DomainName/
+  | ClockN !DomainName
+  -- ^ ClockN type corresponding to domain /DomainName/
   | Reset !DomainName
   -- ^ Reset type corresponding to domain /DomainName/
   | Enable !DomainName
@@ -452,7 +469,7 @@ data HWType
   | CustomProduct !Text !DataRepr' !Size (Maybe [Text]) [(FieldAnn, HWType)]
   -- ^ Same as Product, but with a user specified bit representation. For more
   -- info, see: Clash.Annotations.BitRepresentations.
-  | Annotated [Attr'] !HWType
+  | Annotated [Attr Text] !HWType
   -- ^ Annotated with HDL attributes
   | KnownDomain !DomainName !Integer !ActiveEdge !ResetKind !InitBehavior !ResetPolarity
   -- ^ Domain name, period, active edge, reset kind, initial value behavior
@@ -460,9 +477,17 @@ data HWType
   -- ^ File type for simulation-level I/O
   deriving (Eq, Ord, Show, Generic, NFData, Hashable)
 
+-- | Smart constructor for 'Annotated'. Wraps the given type in an 'Annotated'
+-- if the attribute list is non-empty. If it is empty, it will return the given
+-- 'HWType' unchanged.
+annotated :: [Attr Text] -> HWType -> HWType
+annotated [] t = t
+annotated attrs t = Annotated attrs t
+
 hwTypeDomain :: HWType -> Maybe DomainName
 hwTypeDomain = \case
   Clock dom -> Just dom
+  ClockN dom -> Just dom
   Reset dom -> Just dom
   Enable dom -> Just dom
   KnownDomain dom _ _ _ _ _ -> Just dom
@@ -470,7 +495,7 @@ hwTypeDomain = \case
 
 -- | Extract hardware attributes from Annotated. Returns an empty list if
 -- non-Annotated given or if Annotated has an empty list of attributes.
-hwTypeAttrs :: HWType -> [Attr']
+hwTypeAttrs :: HWType -> [Attr Text]
 hwTypeAttrs (Annotated attrs _type) = attrs
 hwTypeAttrs _                       = []
 
@@ -503,6 +528,7 @@ data Declaration
   -- | Signal assignment
   = Assignment
       !Identifier -- ^ Signal to assign
+      !Usage      -- ^ How the signal is assigned
       !Expr       -- ^ Assigned expression
 
   -- | Conditional signal assignment:
@@ -517,7 +543,7 @@ data Declaration
   | InstDecl
       EntityOrComponent                  -- ^ Whether it's an entity or a component
       (Maybe Text)                       -- ^ Library instance is defined in
-      [Attr']                            -- ^ Attributes to add to the generated code
+      [Attr Text]                        -- ^ Attributes to add to the generated code
       !Identifier                        -- ^ The component's (or entity's) name
       !Identifier                        -- ^ Instance label
       [(Expr,HWType,Expr)]               -- ^ List of parameters for this component (param name, param type, param value)
@@ -532,12 +558,19 @@ data Declaration
       !BlackBox                -- ^ Template tokens
       BlackBoxContext          -- ^ Context in which tokens should be rendered
 
+  -- | @component@ declaration (VHDL).
+  --
+  -- See [this tutorial](https://www.ics.uci.edu/~jmoorkan/vhdlref/compdec.html);
+  -- refer to §4.5 of IEEE 1076-1993
+  | CompDecl
+      !Text
+      [(Text, PortDirection, HWType)]
+
   -- | Signal declaration
   | NetDecl'
       (Maybe Comment)                -- ^ Note; will be inserted as a comment in target hdl
-      WireOrReg                      -- ^ Wire or register
       !Identifier                    -- ^ Name of signal
-      (Either IdentifierText HWType) -- ^ Pointer to type of signal or type of signal
+      HWType                         -- ^ Type of signal
       (Maybe Expr)                   -- ^ Initial value
       -- ^ Signal declaration
 
@@ -578,14 +611,111 @@ data Seq
       [(Maybe Literal,[Seq])]  -- ^ List of: (Maybe match, RHS of Alternative)
   deriving Show
 
+-- | Procedural assignment in HDL can be blocking or non-blocking. This
+-- determines when the assignment takes place in simulation. The name refers to
+-- whether evaluation of the remaining statements in a process is blocked
+-- until the assignment is performed or not.
+--
+-- See Also:
+--
+-- IEEE 1364-2001, sections 9.2.1 and 9.2.2
+-- IEEE 1076-1993, sections 8.4 and 8.5
+--
+data Blocking
+  = NonBlocking
+  -- ^ A non-blocking assignment means the new value is not observed until the
+  -- next time step in simulation. Using the signal later in the process will
+  -- continue to return the old value.
+  | Blocking
+  -- ^ A blocking assignment means the new value is observed immediately. Using
+  -- the signal later in the process will return the new value.
+  deriving (Binary, Eq, Generic, Hashable, NFData, Show)
+
+-- NOTE [`Semigroup` instances for `Blocking` and `Usage`]
+instance Semigroup Blocking where
+  NonBlocking <> y = y
+  Blocking    <> _ = Blocking
+
+-- | The usage of a signal refers to how the signal is written to in netlist.
+-- This is used to determine if the signal should be a @wire@ or @reg@ in
+-- (System)Verilog, or a @signal@ or @variable@ in VHDL.
+--
+data Usage
+  = Cont
+  -- ^ Continuous assignment, which occurs in a concurrent context.
+  | Proc Blocking
+  -- ^ Procedural assignment, which occurs in a sequential context.
+  deriving (Binary, Eq, Generic, Hashable, NFData, Show)
+
+-- NOTE [`Semigroup` instances for `Blocking` and `Usage`]
+instance Semigroup Usage where
+  Cont    <> y      = y
+  Proc x  <> Proc y = Proc (x <> y)
+  Proc x  <> _      = Proc x
+
+{-
+NOTE [`Semigroup` instances for `Blocking` and `Usage`]
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+Usages (and Blocking) are combined by taking the most restrictive usage, where
+most restrictive means "has the most influence over the choice of declaration".
+Clash produces three types of assignment:
+
+  * continuous
+  * prodecural non-blocking
+  * prodecural blocking
+
+Both VHDl and (System)Verilog have a type of declaration which only admits one
+type of assignment. This is the most restrictive for that HDL. However, since
+that would involve knowing the HDL type in these Semigroup instances, the
+most restrictive here is based on ordering where the most restrictive for each
+HDL is an extreme value (max for VHDL, min for Verilog). i.e.
+
+          |-------------------------------------|
+          | Continuous | NonBlocking | Blocking |
+|---------|-------------------------------------|
+|    VHDL |         signal           | variable |
+|---------|-------------------------------------|
+| Verilog |   wire     |          reg           |
+|---------|-------------------------------------|
+-}
+
+instance FromJSON Usage where
+  parseJSON = Aeson.withText "Usage" $ \case
+    "Continuous"  -> pure Cont
+    "NonBlocking" -> pure (Proc NonBlocking)
+    "Blocking"    -> pure (Proc Blocking)
+    str           -> fail $ mconcat
+      [ "Could not parse usage: "
+      , show str
+      , "\nRecognized values are 'Continuous', 'NonBlocking' and 'Blocking'"
+      ]
+
+-- See NOTE [`Text` key for `UsageMap`]
+type UsageMap = Map Text Usage
+
+lookupUsage :: Identifier -> UsageMap -> Maybe Usage
+lookupUsage i = Map.lookup (Id.toText i)
+
+{-
+NOTE [`Text` key for `UsageMap`]
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+We would like to use netlist identifiers as the key for the usage map, since
+conceptually it is a map from an identifier to how it is used in assignments.
+However, in practice we commonly end up with the same textual identifier
+appearing in different ways in the netlist.
+
+The most obvious example of this are identifiers that appear as both
+`UniqueIdentifier` and `RawIdentifier`. If we track the usage on the raw
+identifier, but the `NetDecl` uses the `UniqueIdentifier` then the wrong
+declaration may be used in the rendered HDL.
+
+Attempting to fix this by not generating the same textual identifier in
+different ways proved difficult, so for now the key type is `Text` instead.
+-}
+
 data EntityOrComponent = Entity | Comp | Empty
   deriving Show
 
-data WireOrReg = Wire | Reg
-  deriving (Show,Generic)
-
-instance NFData WireOrReg
-
 pattern NetDecl
   :: Maybe Comment
   -- ^ Note; will be inserted as a comment in target hdl
@@ -594,9 +724,9 @@ pattern NetDecl
   -> HWType
   -- ^ Type of signal
   -> Declaration
-pattern NetDecl note d ty <- NetDecl' note Wire d (Right ty) _
+pattern NetDecl note d ty <- NetDecl' note d ty _
   where
-    NetDecl note d ty = NetDecl' note Wire d (Right ty) Nothing
+    NetDecl note d ty = NetDecl' note d ty Nothing
 
 data PortDirection = In | Out
   deriving (Eq,Ord,Show,Generic,NFData,Hashable)
@@ -606,11 +736,18 @@ instance NFData Declaration where
 
 -- | Expression Modifier
 data Modifier
-  = Indexed (HWType,Int,Int) -- ^ Index the expression: (Type of expression,DataCon tag,Field Tag)
-  | DC (HWType,Int)          -- ^ See expression in a DataCon context: (Type of the expression, DataCon tag)
-  | VecAppend                -- ^ See the expression in the context of a Vector append operation
-  | RTreeAppend              -- ^ See the expression in the context of a Tree append operation
-  | Sliced (HWType,Int,Int)  -- ^ Slice the identifier of the given type from start to end
+  = Indexed (HWType, Int, Int)
+  -- ^ Index the expression: (Type of expression, DataCon tag, Field Tag). Note
+  -- that the type of the expression is the type we are slicing from, not the type
+  -- returned by the index operation.
+  | DC (HWType, Int)
+  -- ^ See expression in a DataCon context: (Type of the expression, DataCon tag)
+  | VecAppend
+  -- ^ See the expression in the context of a Vector append operation
+  | RTreeAppend
+  -- ^ See the expression in the context of a Tree append operation
+  | Sliced (HWType, Int, Int)
+  -- ^ Slice the identifier of the given type from start to end
   | Nested Modifier Modifier
   deriving Show
 
@@ -651,6 +788,24 @@ data Expr
 instance NFData Expr where
   rnf x = x `seq` ()
 
+isConstExpr :: Expr -> Bool
+isConstExpr = \case
+  Literal{} -> True
+  DataCon _ _ es -> all isConstExpr es
+  Identifier{} -> False
+  DataTag{} -> False
+  BlackBoxE nm _ _ _ _ ctx _
+    -- When using SimIO, `reg` creates (in Haskell) the mutable reference to
+    -- some value. The blackbox for this however is simply `~ARG[0]`, so if
+    -- the argument given is constant, the rendered HDL will also be constant.
+    | nm == "Clash.Explicit.SimIO.reg" ->
+        all (\(e, _, _) -> isConstExpr e) (bbInputs ctx)
+    | otherwise -> False
+  ToBv _ _ e -> isConstExpr e
+  FromBv _ _ e -> isConstExpr e
+  IfThenElse{} -> False
+  Noop -> False
+
 -- | Literals used in an expression
 data Literal
   = NumLit    !Integer          -- ^ Number literal
@@ -689,7 +844,7 @@ data BlackBoxContext
   , bbInputs :: [(Expr,HWType,Bool)]
   -- ^ Argument names, types, and whether it is a literal
   , bbFunctions :: IntMap [(Either BlackBox (Identifier,[Declaration])
-                          ,WireOrReg
+                          ,Usage
                           ,[BlackBoxTemplate]
                           ,[BlackBoxTemplate]
                           ,[((Text,Text),BlackBox)]
@@ -758,7 +913,7 @@ data NetlistId
   | MultiId [Id]
   -- ^ A split identifier (into several sub-identifiers), needed to assign
   -- expressions of types that have to be split apart (e.g. tuples of Files)
-  deriving Show
+  deriving (Eq, Show)
 
 -- | Eliminator for 'NetlistId', fails on 'MultiId'
 netlistId1
@@ -799,6 +954,11 @@ data DeclarationType
   = Concurrent
   | Sequential
 
+-- | Default usage for a type of declaration (concurrent or sequential)
+declTypeUsage :: DeclarationType -> Usage
+declTypeUsage Concurrent = Cont
+declTypeUsage Sequential = Proc Blocking
+
 emptyBBContext :: Text -> BlackBoxContext
 emptyBBContext name
   = Context
@@ -839,6 +999,12 @@ class HasIdentifierSet s where
 instance HasIdentifierSet IdentifierSet where
   identifierSet = ($)
 
+instance HasUsageMap NetlistState where
+  usageMap = usages
+
+instance HasIdentifierSet s => HasIdentifierSet (s, a) where
+  identifierSet = Lens._1 . identifierSet
+
 -- | An "IdentifierSetMonad" supports unique name generation for Clash Netlist
 class Monad m => IdentifierSetMonad m where
   identifierSetM :: (IdentifierSet -> IdentifierSet) -> m IdentifierSet
