@@ -9,6 +9,7 @@
   Transformations for compile-time reduction of expressions / primitives.
 -}
 
+{-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE OverloadedStrings #-}
 
 module Clash.Normalize.Transformations.Reduce
@@ -23,25 +24,28 @@ import qualified Data.Either as Either
 import qualified Data.List as List
 import qualified Data.List.Extra as List
 import qualified Data.Maybe as Maybe
+import Data.Maybe (fromMaybe, listToMaybe)
 import GHC.Stack (HasCallStack)
 
 import Clash.Core.FreeVars (typeFreeVars)
 import Clash.Core.HasType
 import Clash.Core.Name (nameOcc)
+import Clash.Core.Pretty (showPpr)
 import Clash.Core.Subst (Subst, extendIdSubst, substTm)
 import Clash.Core.Term
-  ( LetBinding, PrimInfo(..), Term(..), TickInfo(..), collectArgs
-  , collectArgsTicks, mkApps, mkTicks)
+  ( CoreContext(..), LetBinding, PrimInfo(..), Term(..), TickInfo(..), collectArgs
+  , collectArgsTicks, mkApps, mkTicks, mkTmApps)
 import Clash.Core.TyCon (tyConDataCons)
-import Clash.Core.Type (TypeView(..), mkTyConApp, tyView)
-import Clash.Core.Util (mkVec, shouldSplit, tyNatSize)
+import Clash.Core.Type (Type, TypeView(..), mkTyConApp, splitFunForallTy, tyView)
+import Clash.Core.Util (mkVec, shouldSplit, tyNatSize, mkInternalVar)
+import Clash.Core.VarEnv (extendInScopeSet)
+import qualified Clash.Data.UniqMap as UniqMap
 import Clash.Normalize.PrimitiveReductions
 import Clash.Normalize.Primitives (removedArg)
 import Clash.Normalize.Types (NormRewrite, NormalizeSession)
 import Clash.Normalize.Util (shouldReduce)
 import Clash.Rewrite.Types (TransformContext(..), tcCache, normalizeUltra)
 import Clash.Rewrite.Util (changed, isUntranslatableType, setChanged, whnfRW)
-import Clash.Unique (lookupUniqMap)
 
 -- | XXX: is given inverse topologically sorted binders, but returns
 -- topologically sorted binders
@@ -146,24 +150,26 @@ reduceConst _ e = return e
 --
 -- See https://github.com/clash-lang/clash-compiler/issues/1606
 reduceNonRepPrim :: HasCallStack => NormRewrite
-reduceNonRepPrim c@(TransformContext is0 ctx) e@(App _ _) | (Prim p, args, ticks) <- collectArgsTicks e = do
+reduceNonRepPrim c@(TransformContext _ ctx) e@(App _ _) | (Prim p, args, ticks) <- collectArgsTicks e = do
   tcm <- Lens.view tcCache
   ultra <- Lens.view normalizeUltra
   let eTy = inferCoreTypeOf tcm e
-  case tyView eTy of
+  let resTy = snd (splitFunForallTy eTy)
+  case tyView resTy of
     (TyConApp vecTcNm@(nameOcc -> "Clash.Sized.Vector.Vec")
               [runExcept . tyNatSize tcm -> Right 0, aTy]) -> do
-      let (Just vecTc) = lookupUniqMap vecTcNm tcm
-          [nilCon,consCon] = tyConDataCons vecTc
-          nilE = mkVec nilCon consCon aTy 0 []
+      let nilE = fromMaybe (error "reduceNonRepPrim: unable to create Vec DCs") $ do
+            vecTc <- UniqMap.lookup vecTcNm tcm
+            [nilCon,consCon] <- pure (tyConDataCons vecTc)
+            return (mkVec nilCon consCon aTy 0 [])
       changed (mkTicks nilE ticks)
     tv -> let argLen = length args in case primName p of
-      "Clash.Sized.Vector.zipWith" | argLen == 7 -> do
-        let [lhsElTy,rhsElty,resElTy,nTy] = Either.rights args
-            TyConApp vecTcNm _ = tv
-            lhsTy = mkTyConApp vecTcNm [nTy,lhsElTy]
-            rhsTy = mkTyConApp vecTcNm [nTy,rhsElty]
-        case runExcept (tyNatSize tcm nTy) of
+      "Clash.Sized.Vector.zipWith"
+        | (tmArgs,[lhsElTy,rhsElty,resElTy,nTy]) <- Either.partitionEithers args
+        , TyConApp vecTcNm _ <- tv
+        , let lhsTy = mkTyConApp vecTcNm [nTy,lhsElTy]
+        , let rhsTy = mkTyConApp vecTcNm [nTy,rhsElty]
+        -> case runExcept (tyNatSize tcm nTy) of
           Right n -> do
             shouldReduce1 <- List.orM [ pure (ultra || n < 2)
                                  , shouldReduce ctx
@@ -171,18 +177,20 @@ reduceNonRepPrim c@(TransformContext is0 ctx) e@(App _ _) | (Prim p, args, ticks
                                         [lhsElTy,rhsElty,resElTy]
                                  -- Note [Unroll shouldSplit types]
                                  , pure (any (Maybe.isJust . shouldSplit tcm)
-                                             [lhsTy,rhsTy,eTy]) ]
+                                             [lhsTy,rhsTy,resTy]) ]
             if shouldReduce1
-               then let [fun,lhsArg,rhsArg] = Either.lefts args
-                    in  (`mkTicks` ticks) <$>
-                        reduceZipWith c p n lhsElTy rhsElty resElTy fun lhsArg rhsArg
+               then abstractOverMissingArgs ticks tmArgs eTy c
+                      (reduceZipWith p n lhsElTy rhsElty resElTy)
                else return e
           _ -> return e
-      "Clash.Sized.Vector.map" | argLen == 5 -> do
-        let [argElTy,resElTy,nTy] = Either.rights args
-            TyConApp vecTcNm _ = tv
-            argTy = mkTyConApp vecTcNm [nTy,argElTy]
-        case runExcept (tyNatSize tcm nTy) of
+        | argLen >= 4
+        -> error ("reduceNonRepPrim: zipWith bad args" <> showPpr e)
+
+      "Clash.Sized.Vector.map"
+        | (tmArgs,[argElTy,resElTy,nTy]) <- Either.partitionEithers args
+        , TyConApp vecTcNm _ <- tv
+        , let argTy = mkTyConApp vecTcNm [nTy,argElTy]
+        -> case runExcept (tyNatSize tcm nTy) of
           Right n -> do
             shouldReduce1 <- List.orM [ pure (ultra || n < 2 )
                                  , shouldReduce ctx
@@ -190,23 +198,27 @@ reduceNonRepPrim c@(TransformContext is0 ctx) e@(App _ _) | (Prim p, args, ticks
                                         [argElTy,resElTy]
                                  -- Note [Unroll shouldSplit types]
                                  , pure (any (Maybe.isJust . shouldSplit tcm)
-                                             [argTy,eTy]) ]
+                                             [argTy,resTy]) ]
             if shouldReduce1
-               then let [fun,arg] = Either.lefts args
-                    in  (`mkTicks` ticks) <$> reduceMap c p n argElTy resElTy fun arg
+               then abstractOverMissingArgs ticks tmArgs eTy c
+                      (reduceMap p n argElTy resElTy)
                else return e
           _ -> return e
-      "Clash.Sized.Vector.traverse#" | argLen == 7 ->
-        let [aTy,fTy,bTy,nTy] = Either.rights args
-        in  case runExcept (tyNatSize tcm nTy) of
-          Right n ->
-            let [dict,fun,arg] = Either.lefts args
-            in  (`mkTicks` ticks) <$> reduceTraverse c n aTy fTy bTy dict fun arg
+        | argLen >= 3
+        -> error ("reduceNonRepPrim: map bad args" <> showPpr e)
+
+      "Clash.Sized.Vector.traverse#"
+        | (tmArgs,[aTy,fTy,bTy,nTy]) <- Either.partitionEithers args
+        -> case runExcept (tyNatSize tcm nTy) of
+          Right n -> abstractOverMissingArgs ticks tmArgs eTy c (reduceTraverse n aTy fTy bTy)
           _ -> return e
-      "Clash.Sized.Vector.fold" | argLen == 4 -> do
-        let ([fun,arg],[nTy,aTy]) = Either.partitionEithers args
-            argTy = inferCoreTypeOf tcm arg
-        case runExcept (tyNatSize tcm nTy) of
+        | argLen >= 4
+        -> error ("reduceNonRepPrim: traverse# bad args" <> showPpr e)
+
+      "Clash.Sized.Vector.fold"
+        | (tmArgs,[nTy,aTy]) <- Either.partitionEithers args
+        , (_:Right argTy:_) <- fst (splitFunForallTy (piResultTys tcm (primType p) [nTy,aTy]))
+        -> case runExcept (tyNatSize tcm nTy) of
           Right n -> do
             shouldReduce1 <- List.orM [ pure (ultra || n == 0)
                                  , shouldReduce ctx
@@ -214,13 +226,16 @@ reduceNonRepPrim c@(TransformContext is0 ctx) e@(App _ _) | (Prim p, args, ticks
                                  -- Note [Unroll shouldSplit types]
                                  , pure (Maybe.isJust (shouldSplit tcm argTy))]
             if shouldReduce1 then
-              (`mkTicks` ticks) <$> reduceFold c (n + 1) aTy fun arg
+              abstractOverMissingArgs ticks tmArgs eTy c (reduceFold (n + 1) aTy)
             else return e
           _ -> return e
-      "Clash.Sized.Vector.foldr" | argLen == 6 ->
-        let ([fun,start,arg],[aTy,bTy,nTy]) = Either.partitionEithers args
-            argTy = inferCoreTypeOf tcm arg
-        in  case runExcept (tyNatSize tcm nTy) of
+        | argLen >= 2
+        -> error ("reduceNonRepPrim: fold bad args" <> showPpr e)
+
+      "Clash.Sized.Vector.foldr"
+        | (tmArgs,[aTy,bTy,nTy]) <- Either.partitionEithers args
+        , (_:_:Right argTy:_) <- fst (splitFunForallTy (piResultTys tcm (primType p) [aTy,bTy,nTy]))
+        -> case runExcept (tyNatSize tcm nTy) of
           Right n -> do
             shouldReduce1 <- List.orM [ pure ultra
                                  , shouldReduce ctx
@@ -228,63 +243,73 @@ reduceNonRepPrim c@(TransformContext is0 ctx) e@(App _ _) | (Prim p, args, ticks
                                  -- Note [Unroll shouldSplit types]
                                  , pure (Maybe.isJust (shouldSplit tcm argTy)) ]
             if shouldReduce1
-              then (`mkTicks` ticks) <$> reduceFoldr c p n aTy fun start arg
+              then abstractOverMissingArgs ticks tmArgs eTy c (reduceFoldr p n aTy)
               else return e
           _ -> return e
-      "Clash.Sized.Vector.dfold" | argLen == 8 ->
-        let ([_kn,_motive,fun,start,arg],[_mTy,nTy,aTy]) = Either.partitionEithers args
-        in  case runExcept (tyNatSize tcm nTy) of
-          Right n -> (`mkTicks` ticks) <$> reduceDFold is0 n aTy fun start arg
+        | argLen >= 3
+        -> error ("reduceNonRepPrim: foldr bad args" <> showPpr e)
+
+      "Clash.Sized.Vector.dfold"
+        | (tmArgs,[_mTy,nTy,aTy]) <- Either.partitionEithers args
+        -> case runExcept (tyNatSize tcm nTy) of
+          Right n -> abstractOverMissingArgs ticks tmArgs eTy c (reduceDFold n aTy)
           _ -> return e
-      "Clash.Sized.Vector.++" | argLen == 5 ->
-        let [nTy,aTy,mTy] = Either.rights args
-            [lArg,rArg]   = Either.lefts args
-        in case (runExcept (tyNatSize tcm nTy), runExcept (tyNatSize tcm mTy)) of
-              (Right n, Right m)
-                | n == 0 -> changed rArg
-                | m == 0 -> changed lArg
-                | otherwise -> do
-                    shouldReduce1 <- List.orM [ shouldReduce ctx
+        | argLen >= 3
+        -> error ("reduceNonRepPrim: dfold bad args" <> showPpr e)
+
+      "Clash.Sized.Vector.++"
+        | (tmArgs,[nTy,aTy,mTy]) <- Either.partitionEithers args
+        -> case (runExcept (tyNatSize tcm nTy), runExcept (tyNatSize tcm mTy)) of
+              (Right n, Right m) -> do
+                    shouldReduce1 <- List.orM [ pure (n==0)
+                                         , pure (m==0)
+                                         , shouldReduce ctx
                                          , isUntranslatableType_not_poly aTy
                                          -- Note [Unroll shouldSplit types]
-                                         , pure (Maybe.isJust (shouldSplit tcm eTy)) ]
+                                         , pure (Maybe.isJust (shouldSplit tcm resTy)) ]
                     if shouldReduce1
-                       then (`mkTicks` ticks) <$> reduceAppend is0 n m aTy lArg rArg
+                       then abstractOverMissingArgs ticks tmArgs eTy c (reduceAppend n m aTy)
                        else return e
               _ -> return e
-      "Clash.Sized.Vector.head" | argLen == 3 -> do
-        let [nTy,aTy] = Either.rights args
-            [vArg]    = Either.lefts args
-            argTy     = inferCoreTypeOf tcm vArg
-        case runExcept (tyNatSize tcm nTy) of
+        | argLen >= 3
+        -> error ("reduceNonRepPrim: ++ bad args" <> showPpr e)
+
+      "Clash.Sized.Vector.head"
+        | (tmArgs,[nTy,aTy]) <- Either.partitionEithers args
+        , (Right argTy:_) <- fst (splitFunForallTy (piResultTys tcm (primType p) [nTy,aTy]))
+        -> case runExcept (tyNatSize tcm nTy) of
           Right n -> do
             shouldReduce1 <- List.orM [ shouldReduce ctx
                                  , isUntranslatableType_not_poly aTy
                                  -- Note [Unroll shouldSplit types]
                                  , pure (Maybe.isJust (shouldSplit tcm argTy)) ]
             if shouldReduce1
-               then (`mkTicks` ticks) <$> reduceHead is0 (n+1) aTy vArg
+               then abstractOverMissingArgs ticks tmArgs eTy c (reduceHead (n+1) aTy)
                else return e
           _ -> return e
-      "Clash.Sized.Vector.tail" | argLen == 3 -> do
-        let [nTy,aTy] = Either.rights args
-            [vArg]    = Either.lefts args
-            argTy     = inferCoreTypeOf tcm vArg
-        case runExcept (tyNatSize tcm nTy) of
+        | argLen >= 2
+        -> error ("reduceNonRepPrim: head bad args" <> showPpr e)
+
+      "Clash.Sized.Vector.tail"
+        | (tmArgs,[nTy,aTy]) <- Either.partitionEithers args
+        , (Right argTy:_) <- fst (splitFunForallTy (piResultTys tcm (primType p) [nTy,aTy]))
+        -> case runExcept (tyNatSize tcm nTy) of
           Right n -> do
             shouldReduce1 <- List.orM [ shouldReduce ctx
                                  , isUntranslatableType_not_poly aTy
                                  -- Note [Unroll shouldSplit types]
                                  , pure (Maybe.isJust (shouldSplit tcm argTy)) ]
             if shouldReduce1
-               then (`mkTicks` ticks) <$> reduceTail is0 (n+1) aTy vArg
+               then abstractOverMissingArgs ticks tmArgs eTy c (reduceTail (n+1) aTy)
                else return e
           _ -> return e
-      "Clash.Sized.Vector.last" | argLen == 3 -> do
-        let [nTy,aTy] = Either.rights args
-            [vArg]    = Either.lefts args
-            argTy     = inferCoreTypeOf tcm vArg
-        case runExcept (tyNatSize tcm nTy) of
+        | argLen >= 2
+        -> error ("reduceNonRepPrim: tail bad args" <> showPpr e)
+
+      "Clash.Sized.Vector.last"
+        | (tmArgs,[nTy,aTy]) <- Either.partitionEithers args
+        , (Right argTy:_) <- fst (splitFunForallTy (piResultTys tcm (primType p) [nTy,aTy]))
+        -> case runExcept (tyNatSize tcm nTy) of
           Right n -> do
             shouldReduce1 <- List.orM [ shouldReduce ctx
                                  , isUntranslatableType_not_poly aTy
@@ -292,27 +317,32 @@ reduceNonRepPrim c@(TransformContext is0 ctx) e@(App _ _) | (Prim p, args, ticks
                                  , pure (Maybe.isJust (shouldSplit tcm argTy))
                                  ]
             if shouldReduce1
-               then (`mkTicks` ticks) <$> reduceLast is0 (n+1) aTy vArg
+               then abstractOverMissingArgs ticks tmArgs eTy c (reduceLast (n+1) aTy)
                else return e
           _ -> return e
-      "Clash.Sized.Vector.init" | argLen == 3 -> do
-        let [nTy,aTy] = Either.rights args
-            [vArg]    = Either.lefts args
-            argTy     = inferCoreTypeOf tcm vArg
-        case runExcept (tyNatSize tcm nTy) of
+        | argLen >= 2
+        -> error ("reduceNonRepPrim: last bad args" <> showPpr e)
+
+      "Clash.Sized.Vector.init"
+        | (tmArgs,[nTy,aTy]) <- Either.partitionEithers args
+        , (Right argTy:_) <- fst (splitFunForallTy (piResultTys tcm (primType p) [nTy,aTy]))
+        -> case runExcept (tyNatSize tcm nTy) of
           Right n -> do
             shouldReduce1 <- List.orM [ shouldReduce ctx
                                  , isUntranslatableType_not_poly aTy
                                  -- Note [Unroll shouldSplit types]
                                  , pure (Maybe.isJust (shouldSplit tcm argTy)) ]
             if shouldReduce1
-               then (`mkTicks` ticks) <$> reduceInit is0 p n aTy vArg
+               then abstractOverMissingArgs ticks tmArgs eTy c (reduceInit p n aTy)
                else return e
           _ -> return e
-      "Clash.Sized.Vector.unconcat" | argLen == 6 -> do
-        let ([_knN,sm,arg],[nTy,mTy,aTy]) = Either.partitionEithers args
-            argTy = inferCoreTypeOf tcm arg
-        case (runExcept (tyNatSize tcm nTy), runExcept (tyNatSize tcm mTy)) of
+        | argLen >= 2
+        -> error ("reduceNonRepPrim: init bad args" <> showPpr e)
+
+      "Clash.Sized.Vector.unconcat"
+        | (tmArgs,[nTy,mTy,aTy]) <- Either.partitionEithers args
+        , (_:_:Right argTy:_) <- fst (splitFunForallTy (piResultTys tcm (primType p) [nTy,mTy,aTy]))
+        -> case (runExcept (tyNatSize tcm nTy), runExcept (tyNatSize tcm mTy)) of
           (Right n, Right m) -> do
             shouldReduce1 <- List.orM [ pure (m==0)
                                       , shouldReduce ctx
@@ -321,48 +351,59 @@ reduceNonRepPrim c@(TransformContext is0 ctx) e@(App _ _) | (Prim p, args, ticks
                                       , pure (Maybe.isJust (shouldSplit tcm argTy))
                                       ]
             if shouldReduce1 then
-              (`mkTicks` ticks) <$> reduceUnconcat is0 p n m aTy sm arg
+              abstractOverMissingArgs ticks tmArgs eTy c (reduceUnconcat p n m aTy)
             else
               return e
           _ -> return e
-      "Clash.Sized.Vector.transpose" | argLen == 5 -> do
-        let ([_knN,arg],[mTy,nTy,aTy]) = Either.partitionEithers args
-        case (runExcept (tyNatSize tcm nTy), runExcept (tyNatSize tcm mTy)) of
-          (Right n, Right 0) -> (`mkTicks` ticks) <$> reduceTranspose n 0 aTy arg
+        | argLen >= 3
+        -> error ("reduceNonRepPrim: unconcat bad args" <> showPpr e)
+
+      "Clash.Sized.Vector.transpose"
+        | (tmArgs,[mTy,nTy,aTy]) <- Either.partitionEithers args
+        -> case (runExcept (tyNatSize tcm nTy), runExcept (tyNatSize tcm mTy)) of
+          (Right n, Right 0) -> abstractOverMissingArgs ticks tmArgs eTy c (reduceTranspose n 0 aTy)
           _ -> return e
-      "Clash.Sized.Vector.replicate" | argLen == 4 -> do
-        let ([_sArg,vArg],[nTy,aTy]) = Either.partitionEithers args
-        case runExcept (tyNatSize tcm nTy) of
+        | argLen >= 3
+        -> error ("reduceNonRepPrim: transpose bad args" <> showPpr e)
+
+      "Clash.Sized.Vector.replicate"
+        | (tmArgs,[nTy,aTy]) <- Either.partitionEithers args
+        -> case runExcept (tyNatSize tcm nTy) of
           Right n -> do
             shouldReduce1 <- List.orM [ shouldReduce ctx
                                  , isUntranslatableType_not_poly aTy
                                  -- Note [Unroll shouldSplit types]
-                                 , pure (Maybe.isJust (shouldSplit tcm eTy))
+                                 , pure (Maybe.isJust (shouldSplit tcm resTy))
                                  ]
             if shouldReduce1
-               then (`mkTicks` ticks) <$> reduceReplicate n aTy eTy vArg
+               then abstractOverMissingArgs ticks tmArgs eTy c (reduceReplicate n aTy resTy)
                else return e
           _ -> return e
+        | argLen >= 2
+        -> error ("reduceNonRepPrim: replicate bad args" <> showPpr e)
+
        -- replace_int :: KnownNat n => Vec n a -> Int -> a -> Vec n a
-      "Clash.Sized.Vector.replace_int" | argLen == 6 -> do
-        let ([_knArg,vArg,iArg,aArg],[nTy,aTy]) = Either.partitionEithers args
-        case runExcept (tyNatSize tcm nTy) of
+      "Clash.Sized.Vector.replace_int"
+        | (tmArgs,[nTy,aTy]) <- Either.partitionEithers args
+        -> case runExcept (tyNatSize tcm nTy) of
           Right n -> do
             shouldReduce1 <- List.orM [ pure ultra
                                  , shouldReduce ctx
                                  , isUntranslatableType_not_poly aTy
                                  -- Note [Unroll shouldSplit types]
-                                 , pure (Maybe.isJust (shouldSplit tcm eTy))
+                                 , pure (Maybe.isJust (shouldSplit tcm resTy))
                                  ]
             if shouldReduce1
-               then (`mkTicks` ticks) <$> reduceReplace_int is0 n aTy eTy vArg iArg aArg
+               then abstractOverMissingArgs ticks tmArgs eTy c (reduceReplace_int n aTy resTy)
                else return e
           _ -> return e
+        | argLen >= 2
+        -> error ("reduceNonRepPrim: replace_int bad args" <> showPpr e)
 
-      "Clash.Sized.Vector.index_int" | argLen == 5 -> do
-        let ([_knArg,vArg,iArg],[nTy,aTy]) = Either.partitionEithers args
-            argTy = inferCoreTypeOf tcm vArg
-        case runExcept (tyNatSize tcm nTy) of
+      "Clash.Sized.Vector.index_int"
+        | (tmArgs,[nTy,aTy]) <- Either.partitionEithers args
+        , (_:Right argTy:_) <- fst (splitFunForallTy (piResultTys tcm (primType p) [nTy,aTy]))
+        -> case runExcept (tyNatSize tcm nTy) of
           Right n -> do
             shouldReduce1 <- List.orM [ pure ultra
                                  , shouldReduce ctx
@@ -370,104 +411,122 @@ reduceNonRepPrim c@(TransformContext is0 ctx) e@(App _ _) | (Prim p, args, ticks
                                  -- Note [Unroll shouldSplit types]
                                  , pure (Maybe.isJust (shouldSplit tcm argTy)) ]
             if shouldReduce1
-               then (`mkTicks` ticks) <$> reduceIndex_int is0 n aTy vArg iArg
+               then abstractOverMissingArgs ticks tmArgs eTy c (reduceIndex_int n aTy)
                else return e
           _ -> return e
+        | argLen >= 2
+        -> error ("reduceNonRepPrim: index_int bad args" <> showPpr e)
 
-      "Clash.Sized.Vector.imap" | argLen == 6 -> do
-        let [nTy,argElTy,resElTy] = Either.rights args
-            TyConApp vecTcNm _ = tv
-            argTy = mkTyConApp vecTcNm [nTy,argElTy]
-        case runExcept (tyNatSize tcm nTy) of
+      "Clash.Sized.Vector.imap"
+        | (tmArgs,[nTy,argElTy,resElTy]) <- Either.partitionEithers args
+        , TyConApp vecTcNm _ <- tv
+        , let argTy = mkTyConApp vecTcNm [nTy,argElTy]
+        -> case runExcept (tyNatSize tcm nTy) of
           Right n -> do
             shouldReduce1 <- List.orM [ pure (ultra || n < 2)
                                  , shouldReduce ctx
                                  , List.anyM isUntranslatableType_not_poly [argElTy,resElTy]
                                  -- Note [Unroll shouldSplit types]
                                  , pure (any (Maybe.isJust . shouldSplit tcm)
-                                             [argTy,eTy]) ]
+                                             [argTy,resTy]) ]
             if shouldReduce1
-               then let [_,fun,arg] = Either.lefts args
-                    in  (`mkTicks` ticks) <$> reduceImap c n argElTy resElTy fun arg
+               then abstractOverMissingArgs ticks tmArgs eTy c (reduceImap n argElTy resElTy)
                else return e
           _ -> return e
-      "Clash.Sized.Vector.iterateI" | argLen == 5 ->
-        let ([_kn,f,a],[nTy,aTy]) = Either.partitionEithers args in
-        case runExcept (tyNatSize tcm nTy) of
+        | argLen >= 3
+        -> error ("reduceNonRepPrim: imap bad args" <> showPpr e)
+
+      "Clash.Sized.Vector.iterateI"
+        | (tmArgs,[nTy,aTy]) <- Either.partitionEithers args
+        -> case runExcept (tyNatSize tcm nTy) of
           Right n -> do
             shouldReduce1 <- List.orM
               [ pure (ultra || n < 2)
               , shouldReduce ctx
               , isUntranslatableType_not_poly aTy
               -- Note [Unroll shouldSplit types]
-              , pure (Maybe.isJust (shouldSplit tcm eTy)) ]
+              , pure (Maybe.isJust (shouldSplit tcm resTy)) ]
 
             if shouldReduce1 then
-              (`mkTicks` ticks) <$> reduceIterateI c n aTy eTy f a
+              abstractOverMissingArgs ticks tmArgs eTy c (reduceIterateI n aTy resTy)
             else
               return e
           _ -> return e
-      "Clash.Sized.Vector.dtfold" | argLen == 8 ->
-        let ([_kn,_motive,lrFun,brFun,arg],[_mTy,nTy,aTy]) = Either.partitionEithers args
-        in  case runExcept (tyNatSize tcm nTy) of
-          Right n -> (`mkTicks` ticks) <$> reduceDTFold is0 n aTy lrFun brFun arg
+        | argLen >= 2
+        -> error ("reduceNonRepPrim: iterateI bad args" <> showPpr e)
+
+      "Clash.Sized.Vector.dtfold"
+        | (tmArgs,[_mTy,nTy,aTy]) <- Either.partitionEithers args
+        -> case runExcept (tyNatSize tcm nTy) of
+          Right n -> abstractOverMissingArgs ticks tmArgs eTy c (reduceDTFold n aTy)
           _ -> return e
+        | argLen >= 3
+        -> error ("reduceNonRepPrim: dtfold bad args" <> showPpr e)
 
       "Clash.Sized.Vector.reverse"
         | ultra
-        , ([vArg],[nTy,aTy]) <- Either.partitionEithers args
+        , (tmArgs,[nTy,aTy]) <- Either.partitionEithers args
         , Right n <- runExcept (tyNatSize tcm nTy)
-        -> (`mkTicks` ticks) <$> reduceReverse is0 n aTy vArg
+        -> abstractOverMissingArgs ticks tmArgs eTy c (reduceReverse n aTy)
 
-      "Clash.Sized.RTree.tdfold" | argLen == 8 ->
-        let ([_kn,_motive,lrFun,brFun,arg],[_mTy,nTy,aTy]) = Either.partitionEithers args
-        in  case runExcept (tyNatSize tcm nTy) of
-          Right n -> (`mkTicks` ticks) <$> reduceTFold is0 n aTy lrFun brFun arg
+      "Clash.Sized.RTree.tdfold"
+        | (tmArgs,[_mTy,nTy,aTy]) <- Either.partitionEithers args
+        -> case runExcept (tyNatSize tcm nTy) of
+          Right n -> abstractOverMissingArgs ticks tmArgs eTy c (reduceTFold n aTy)
           _ -> return e
-      "Clash.Sized.RTree.treplicate" | argLen == 4 -> do
-        let ([_sArg,vArg],[nTy,aTy]) = Either.partitionEithers args
+        | argLen >= 3
+        -> error ("reduceNonRepPrim: tdfold bad args" <> showPpr e)
+      "Clash.Sized.RTree.treplicate"
+        | (tmArgs,[nTy,aTy]) <- Either.partitionEithers args ->
         case runExcept (tyNatSize tcm nTy) of
           Right n -> do
             shouldReduce1 <- List.orM [ shouldReduce ctx
                                  , isUntranslatableType False aTy ]
             if shouldReduce1
-               then (`mkTicks` ticks) <$> reduceTReplicate n aTy eTy vArg
+               then abstractOverMissingArgs ticks tmArgs eTy c (reduceTReplicate n aTy resTy)
                else return e
           _ -> return e
-      "Clash.Sized.Internal.BitVector.split#" | argLen == 4 -> do
-        let ([_knArg,bvArg],[nTy,mTy]) = Either.partitionEithers args
+        | argLen >= 2
+        -> error ("reduceNonRepPrim: treplicate bad args" <> showPpr e)
+      "Clash.Sized.Internal.BitVector.split#"
+        | (tmArgs,[nTy,mTy]) <- Either.partitionEithers args ->
         case (runExcept (tyNatSize tcm nTy), runExcept (tyNatSize tcm mTy), tv) of
           (Right n, Right m, TyConApp tupTcNm [lTy,rTy])
-            | n == 0 -> do
-              let (Just tupTc) = lookupUniqMap tupTcNm tcm
-                  [tupDc]      = tyConDataCons tupTc
-                  tup          = mkApps (Data tupDc)
-                                    [Right lTy
-                                    ,Right rTy
-                                    ,Left  bvArg
-                                    ,Left  (TyApp (Prim removedArg) rTy)
-                                    ]
+            | n == 0 -> abstractOverMissingArgs ticks tmArgs eTy c $ \(_kn :: Term) bvArg (_ctx :: TransformContext) -> do
+              let tup = mkApps (Data tupDc)
+                           [Right lTy
+                           ,Right rTy
+                           ,Left  bvArg
+                           ,Left  (TyApp (Prim removedArg) rTy)
+                           ]
 
-              changed (mkTicks tup ticks)
-            | m == 0 -> do
-              let (Just tupTc) = lookupUniqMap tupTcNm tcm
-                  [tupDc]      = tyConDataCons tupTc
-                  tup          = mkApps (Data tupDc)
-                                    [Right lTy
-                                    ,Right rTy
-                                    ,Left  (TyApp (Prim removedArg) lTy)
-                                    ,Left  bvArg
-                                    ]
+              (changed (mkTicks tup ticks) :: NormalizeSession Term)
+            | m == 0 -> abstractOverMissingArgs ticks tmArgs eTy c $ \(_kn :: Term) bvArg (_ctx :: TransformContext) -> do
+              let tup = mkApps (Data tupDc)
+                           [Right lTy
+                           ,Right rTy
+                           ,Left  (TyApp (Prim removedArg) lTy)
+                           ,Left  bvArg
+                           ]
 
-              changed (mkTicks tup ticks)
+              (changed (mkTicks tup ticks) :: NormalizeSession Term)
+           where
+            tupDc = fromMaybe (error "reduceNonRepPrim: faield to create tup DC") $ do
+                    tupTc <- UniqMap.lookup tupTcNm tcm
+                    listToMaybe (tyConDataCons tupTc)
           _ -> return e
+        | argLen >= 3
+        -> error ("reduceNonRepPrim: split# bad args" <> showPpr e)
       "Clash.Sized.Internal.BitVector.eq#"
-        | ([_,_],[nTy]) <- Either.partitionEithers args
+        | (tmArgs,[nTy]) <- Either.partitionEithers args
         , Right 0 <- runExcept (tyNatSize tcm nTy)
         , TyConApp boolTcNm [] <- tv
-        -> let (Just boolTc) = lookupUniqMap boolTcNm tcm
-               [_falseDc,trueDc] = tyConDataCons boolTc
-           in  changed (mkTicks (Data trueDc) ticks)
+        -> abstractOverMissingArgs ticks tmArgs eTy c $ \(_kn :: Term) (_l :: Term) (_r :: Term) (_ctx :: TransformContext) -> do
+           let trueDc = fromMaybe (error "reduceNonRepPrim: failed to create True DC") $ do
+                  boolTc <- UniqMap.lookup boolTcNm tcm
+                  [_falseDc,dc] <- pure (tyConDataCons boolTc)
+                  return dc
+            in (changed (Data trueDc) :: NormalizeSession Term)
       _ -> return e
   where
     isUntranslatableType_not_poly t = do
@@ -478,3 +537,29 @@ reduceNonRepPrim c@(TransformContext is0 ctx) e@(App _ _) | (Prim p, args, ticks
 
 reduceNonRepPrim _ e = return e
 {-# SCC reduceNonRepPrim #-}
+
+class AbstractOverMissingArgs a where
+  -- | Abstract over a primitive until it is saturated
+  abstractOverMissingArgs ::
+    HasCallStack =>
+    -- | Ticks originally tagged to the applied primitive
+    [TickInfo] ->
+    -- | Available arguments
+    [Term] ->
+    -- | The type of the expression containing the applied primitive
+    Type ->
+    -- | The context in which reduceNonRepPrim was called
+    TransformContext ->
+    a ->
+    NormalizeSession Term
+
+instance AbstractOverMissingArgs (TransformContext -> NormalizeSession Term) where
+  abstractOverMissingArgs ticks args _ is f = (`mkTmApps` args) <$> (`mkTicks` ticks) <$> f is
+
+instance AbstractOverMissingArgs a => AbstractOverMissingArgs (Term -> a) where
+  abstractOverMissingArgs ticks (t:ts) ty ctx f = abstractOverMissingArgs ticks ts ty ctx (f t)
+  abstractOverMissingArgs ticks []     (tyView -> FunTy argTy resTy) (TransformContext is0 ctx) f = do
+     newId <- mkInternalVar is0 "arg" argTy
+     let ctx1 = TransformContext (extendInScopeSet is0 newId) (LamBody newId : ctx)
+     Lam newId <$> abstractOverMissingArgs ticks [] resTy ctx1 (f (Var newId))
+  abstractOverMissingArgs _ _ ty _ _ = error ("not a funty: " <> showPpr ty)
