@@ -1,6 +1,6 @@
 {-|
   Copyright   :  (C) 2019,      Myrtle Software Ltd.
-                     2020-2023, QBayLogic B.V.
+                     2020-2024, QBayLogic B.V.
                      2021,      Myrtle.ai
                      2022-2023, Google Inc
   License     :  BSD2 (see the file LICENSE)
@@ -127,7 +127,8 @@ import           Clash.Netlist.Types             hiding (Component, toBit)
 import           Clash.Netlist.Util
 import           Clash.Util                      (clogBase)
 import qualified Data.String.Interpolate         as I
-import           Language.Haskell.TH             (Name)
+import qualified Language.Haskell.TH             as TH
+import qualified Language.Haskell.TH.Syntax      as TH
 import           Prelude
 
 -- | Options for 'blackBoxHaskell' function. Use 'def' from package
@@ -164,9 +165,9 @@ instance Default BlackBoxHaskellOpts where
 --
 -- @[1,2]@ would mean this blackbox __ignores__ its second and third argument.
 blackBoxHaskell
-  :: Name
+  :: TH.Name
   -- ^ blackbox name
-  -> Name
+  -> TH.Name
   -- ^ template function name
   -> BlackBoxHaskellOpts
   -- ^ Options, see data structure for more information
@@ -649,14 +650,32 @@ constructProduct ty els =
 
 -- | Create an n-tuple of 'TExpr'
 tuple :: HasCallStack => [TExpr] -> TExpr
-tuple [] = error $ "nTuple: Cannot create empty tuple"
+tuple [] = error $ "tuple: Cannot create empty tuple"
 tuple [_] =
   -- If we don't put this in: tuple . untuple /= id
-  error $ "nTuple: Cannot create 1-tuple"
+  error $ "tuple: Cannot create 1-tuple"
 tuple els = constructProduct tupTy els
  where
+#if MIN_VERSION_base(4,19,0)
+  tupTy = Product (tupModule <> ".Tuple" <> showt (length els)) Nothing (map ety els)
+  tupModule =
+    $(
+    let tupNm = ''(,)
+    in case (TH.nameModule tupNm, TH.nameBase tupNm) of
+        (Just modNm, "Tuple2") -> TH.lift modNm :: TH.ExpQ
+        _ -> error $ "tuple: (,) has an unexpected name: " <> show tupNm
+    )
+#else
   commas = Text.replicate (length els - 1) ","
-  tupTy = Product ("GHC.Tuple.(" <> commas <> ")") Nothing (map ety els)
+  tupTy = Product (tupModule <> ".(" <> commas <> ")") Nothing (map ety els)
+  tupModule =
+    $(
+    let tupNm = ''(,)
+    in case (TH.nameModule tupNm, TH.nameBase tupNm) of
+        (Just modNm, "(,)") -> TH.lift modNm :: TH.ExpQ
+        _ -> error $ "tuple: (,) has an unexpected name: " <> show tupNm
+    )
+#endif
 
 -- | Try to get the literal string value of an expression.
 getStr :: TExpr -> Maybe String
@@ -842,14 +861,12 @@ instHO bbCtx fPos (resTy, bbResTy) argsWithTypes = do
 --
 -- A typical result is that a
 --
--- @
--- component fifo port
---    ( rst : in std_logic
---    ...
---    ; full : out std_logic
---    ; empty : out std_logic );
---  end component;
--- @
+-- > component fifo port
+-- >    ( rst : in std_logic
+-- >    ...
+-- >    ; full : out std_logic
+-- >    ; empty : out std_logic );
+-- >  end component;
 --
 -- declaration would be added in the appropriate place.
 compInBlock
@@ -1068,9 +1085,7 @@ notExpr nm aExpr = do
 
 -- | Creates a BV that produces the following vhdl:
 --
--- @
---    (0 to n => ARG)
--- @
+-- > (0 to n => ARG)
 --
 -- TODO: Implement for (System)Verilog
 pureToBV
@@ -1090,9 +1105,7 @@ pureToBV nm n arg = do
 
 -- | Creates a BV that produces the following vhdl:
 --
--- @
---    std_logic_vector(resize(ARG, n))
--- @
+-- > std_logic_vector(resize(ARG, n))
 --
 -- TODO: Implement for (System)Verilog
 pureToBVResized
