@@ -1,7 +1,7 @@
 {-|
   Copyright   :  (C) 2012-2016, University of Twente
   License     :  BSD2 (see the file LICENSE)
-  Maintainer  :  Christiaan Baaij <christiaan.baaij@gmail.com>
+  Maintainer  :  QBayLogic B.V. <devops@qbaylogic.com>
 
   Assortment of utility function used in the Clash library
 -}
@@ -88,6 +88,10 @@ instance Show ClashException where
 
 instance Exception.Exception ClashException
 
+-- | Construct a string pattern match out of the given @TemplateHaskell@ name
+namePat :: TH.Name -> TH.Q TH.Pat
+namePat = return . TH.LitP . TH.StringL . show
+
 assertPanic
   :: String -> Int -> a
 assertPanic file ln = Exception.throw
@@ -153,9 +157,9 @@ pprDebugAndThen cont heading prettyMsg =
 -- | A class that can generate unique numbers
 class Monad m => MonadUnique m where
   -- | Get a new unique
-  getUniqueM :: m Int
+  getUniqueM :: m Unique
 
-instance Monad m => MonadUnique (StateT Int m) where
+instance Monad m => MonadUnique (StateT Unique m) where
   getUniqueM = do
     supply <- State.get
     State.modify (+1)
